from nmigen import *

__all__ = ["BlinkDebug"]


class BlinkDebug(Elaboratable):
    def __init__(self, led, divider=20, max_value=8):
        self.led = led
        self.max_value = max_value
        self.divider = divider

        self.value = Signal(range(max_value))

    def elaborate(self, platform):
        m = Module()

        div_counter = Signal(self.divider)
        m.d.sync += div_counter.eq(div_counter + 1)

        def next_div(m, next_state):
            with m.If(div_counter == 0):
                m.next = next_state

        with m.FSM():
            for i in range(self.max_value):
                with m.State("ON_{}".format(i)):
                    m.d.comb += self.led.eq(1)
                    next_div(m, "OFF_{}".format(i))

                with m.State("OFF_{}".format(i)):
                    m.d.comb += self.led.eq(0)
                    if i != self.max_value - 1:
                        with m.If(self.value > i):
                            next_div(m, "ON_{}".format(i + 1))
                        with m.Else():
                            next_div(m, "IDLE_0")
                    else:
                        next_div(m, "IDLE_0")

            idle_states = 4
            for i in range(idle_states):
                with m.State("IDLE_{}".format(i)):
                    m.d.comb += self.led.eq(0)
                    if i != idle_states - 1:
                        next_div(m, "IDLE_{}".format(i + 1))
                    else:
                        next_div(m, "ON_0")

        return m



from nmigen import *
from nmigen.lib.cdc import FFSynchronizer

from nap import StatusSignal, ControlSignal, iterator_with_if_elif
from nap.vendor.xilinx_s7 import Mmcm
from nap.vendor.xilinx_s7.io import _ISerdes


class HispiPhy(Elaboratable):
    def __init__(self, num_lanes=4, bits=12, hispi_domain="hispi"):
        assert bits == 12
        self.hispi_domain = hispi_domain

        self.hispi_clk = Signal()
        self.hispi_lanes = Signal(num_lanes)

        self.bitslip = [Signal() for _ in range(num_lanes)]
        self.out = [Signal(12) for _ in range(num_lanes)]

        self.hispi_x6_in_domain_counter = StatusSignal(32)
        self.enable_bitslip = ControlSignal(reset=1)
        self.word_reverse = ControlSignal()

    def elaborate(self, platform):
        m = Module()

        hispi_6_in = "{}_x6_in".format(self.hispi_domain)
        m.domains += ClockDomain(hispi_6_in)
        m.d.comb += ClockSignal(hispi_6_in).eq(self.hispi_clk)
        m.d[hispi_6_in] += self.hispi_x6_in_domain_counter.eq(self.hispi_x6_in_domain_counter + 1)

        mul = 3
        pll = m.submodules.pll = Mmcm(300e6, mul, 1, input_domain=hispi_6_in.format(self.hispi_domain))
        pll.output_domain("{}_x6".format(self.hispi_domain), mul * 1)
        pll.output_domain("{}_x3".format(self.hispi_domain), mul * 2)
        pll.output_domain("{}_x2".format(self.hispi_domain), mul * 3)
        pll.output_domain("{}".format(self.hispi_domain), mul * 6)

        for lane in range(0, len(self.hispi_lanes)):
            iserdes = m.submodules["hispi_iserdes_" + str(lane)] = _ISerdes(
                data_width=6,
                data_rate="DDR",
                serdes_mode="master",
                interface_type="networking",
                num_ce=1,
                iobDelay="none",
            )

            m.d.comb += iserdes.d.eq(self.hispi_lanes[lane])
            m.d.comb += iserdes.ce[1].eq(1)
            m.d.comb += iserdes.clk.eq(ClockSignal("{}_x6".format(self.hispi_domain)))
            m.d.comb += iserdes.clkb.eq(~ClockSignal("{}_x6".format(self.hispi_domain)))
            m.d.comb += iserdes.rst.eq(ResetSignal("{}_x6".format(self.hispi_domain)))
            m.d.comb += iserdes.clkdiv.eq(ClockSignal("{}_x2".format(self.hispi_domain)))

            data = Signal(12)
            iserdes_output = Signal(6)
            m.d.comb += iserdes_output.eq(Cat(iserdes.q[j] for j in range(1, 7)))

            hispi_x2 = "{}_x2".format(self.hispi_domain)
            lower_upper_half = Signal()
            m.d[hispi_x2] += lower_upper_half.eq(~lower_upper_half)
            with m.If(lower_upper_half):
                m.d[hispi_x2] += data[6:12].eq(iserdes_output)
            with m.Else():
                m.d[hispi_x2] += data[0:6].eq(iserdes_output)

            data_in_hispi_domain = Signal(12)
            m.submodules["data_cdc_{}".format(lane)] = FFSynchronizer(data, data_in_hispi_domain, o_domain=self.hispi_domain)

            hispi_domain = m.d[self.hispi_domain]
            bitslip = Signal()
            was_bitslip = Signal()
            hispi_domain += was_bitslip.eq(bitslip)
            with m.If(self.bitslip[lane] & ~was_bitslip & self.enable_bitslip):
                hispi_domain += bitslip.eq(1)
            with m.Else():
                hispi_domain += bitslip.eq(0)

            serdes_or_emulated_bitslip = Signal()
            with m.If(bitslip):
                hispi_domain += serdes_or_emulated_bitslip.eq(~serdes_or_emulated_bitslip)

            m.d.comb += iserdes.bitslip.eq(bitslip & serdes_or_emulated_bitslip)

            data_order_index = Signal(range(4))
            with m.If(bitslip & ~serdes_or_emulated_bitslip):
                hispi_domain += data_order_index.eq(data_order_index + 1)

            data_order = StatusSignal(range(16))
            setattr(self, "data_order_{}".format(lane), data_order)
            m.d.comb += data_order.eq(Array((1, 4, 9, 12))[data_order_index])

            current = Signal(12)
            last = Signal(12)
            m.d.comb += current.eq(data_in_hispi_domain)
            hispi_domain += last.eq(data_in_hispi_domain)
            reordered = Signal(12)
            parts = [current[0:6], current[6:12], last[0:6], last[6:12]]
            for cond, i in iterator_with_if_elif(range(16), m):
                with cond(data_order == i):
                    first = parts[i % 4]
                    second = parts[i // 4]
                    m.d.comb += reordered.eq(Cat(first, second))

            with m.If(self.word_reverse):
                m.d.comb += self.out[lane].eq(Cat(reordered[i] for i in range(12)))
            with m.Else():
                m.d.comb += self.out[lane].eq(Cat(reordered[i] for i in reversed(range(12))))

            out_status_signal = StatusSignal(12, name="out_{}".format(lane))
            setattr(self, "out_{}".format(lane), out_status_signal)
            m.d.comb += out_status_signal.eq(data_in_hispi_domain)

        return m



# -*- coding: utf-8 -*-

"""
Portions Copyright (c) 2014 CiiNOW Inc.
Written by Alex Izvorski <aizvorski@gmail.com>, <alex@ciinow.com>
2014-03-03 Ported from matlab to python/numpy/scipy
2014-03-04 Added utility functions to read/compare images and video
"""

"""
-----------COPYRIGHT NOTICE STARTS WITH THIS LINE------------
Copyright (c) 2005 The University of Texas at Austin
All rights reserved.
Permission is hereby granted, without written agreement and without license or royalty fees, to use, copy, 
modify, and distribute this code (the source files) and its documentation for
any purpose, provided that the copyright notice in its entirety appear in all copies of this code, and the 
original source of this code, Laboratory for Image and Video Engineering (LIVE, http://live.ece.utexas.edu)
at the University of Texas at Austin (UT Austin, 
http://www.utexas.edu), is acknowledged in any publication that reports research using this code. The research
is to be cited in the bibliography as:
H. R. Sheikh and A. C. Bovik, "Image Information and Visual Quality", IEEE Transactions on 
Image Processing, (to appear).
IN NO EVENT SHALL THE UNIVERSITY OF TEXAS AT AUSTIN BE LIABLE TO ANY PARTY FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, 
OR CONSEQUENTIAL DAMAGES ARISING OUT OF THE USE OF THIS DATABASE AND ITS DOCUMENTATION, EVEN IF THE UNIVERSITY OF TEXAS
AT AUSTIN HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
THE UNIVERSITY OF TEXAS AT AUSTIN SPECIFICALLY DISCLAIMS ANY WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED 
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE. THE DATABASE PROVIDED HEREUNDER IS ON AN "AS IS" BASIS,
AND THE UNIVERSITY OF TEXAS AT AUSTIN HAS NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
-----------COPYRIGHT NOTICE ENDS WITH THIS LINE------------
This software release consists of a MULTISCALE PIXEL DOMAIN, SCALAR GSM implementation of the algorithm described in the paper:
H. R. Sheikh and A. C. Bovik, "Image Information and Visual Quality"., IEEE Transactions on Image Processing, (to appear).
Download manuscript draft from http://live.ece.utexas.edu in the Publications link.
THE PIXEL DOMAIN ALGORITHM IS NOT DESCRIBED IN THE PAPER. THIS IS A COMPUTATIONALLY SIMPLER
DERIVATIVE OF THE ALGORITHM PRESENTED IN THE PAPER
Input : (1) img1: The reference image as a matrix
        (2) img2: The distorted image (order is important)
Output: (1) VIF the visual information fidelity measure between the two images
Default Usage:
   Given 2 test images img1 and img2, whose dynamic range is 0-255
   vif = vifvec(img1, img2);
Advanced Usage:
   Users may want to modify the parameters in the code. 
   (1) Modify sigma_nsq to find tune for your image dataset.
Email comments and bug reports to hamid.sheikh@ieee.org
"""

import numpy
import scipy.signal
import scipy.ndimage


def vifp_mscale(ref, dist):
    sigma_nsq = 2
    eps = 1e-10

    num = 0.0
    den = 0.0
    for scale in range(1, 5):

        N = 2 ** (4 - scale + 1) + 1
        sd = N / 5.0

        if (scale > 1):
            ref = scipy.ndimage.gaussian_filter(ref, sd)
            dist = scipy.ndimage.gaussian_filter(dist, sd)
            ref = ref[::2, ::2]
            dist = dist[::2, ::2]

        mu1 = scipy.ndimage.gaussian_filter(ref, sd)
        mu2 = scipy.ndimage.gaussian_filter(dist, sd)
        mu1_sq = mu1 * mu1
        mu2_sq = mu2 * mu2
        mu1_mu2 = mu1 * mu2
        sigma1_sq = scipy.ndimage.gaussian_filter(ref * ref, sd) - mu1_sq
        sigma2_sq = scipy.ndimage.gaussian_filter(dist * dist, sd) - mu2_sq
        sigma12 = scipy.ndimage.gaussian_filter(ref * dist, sd) - mu1_mu2

        sigma1_sq[sigma1_sq < 0] = 0
        sigma2_sq[sigma2_sq < 0] = 0

        g = sigma12 / (sigma1_sq + eps)
        sv_sq = sigma2_sq - g * sigma12

        g[sigma1_sq < eps] = 0
        sv_sq[sigma1_sq < eps] = sigma2_sq[sigma1_sq < eps]
        sigma1_sq[sigma1_sq < eps] = 0

        g[sigma2_sq < eps] = 0
        sv_sq[sigma2_sq < eps] = 0

        sv_sq[g < 0] = sigma2_sq[g < 0]
        g[g < 0] = 0
        sv_sq[sv_sq <= eps] = eps

        num += numpy.sum(numpy.log10(1 + g * g * sigma1_sq / (sv_sq + sigma_nsq)))
        den += numpy.sum(numpy.log10(1 + sigma1_sq / sigma_nsq))

    vifp = num / den

    if numpy.isnan(vifp):
        return 1.0
    else:
        return vifp



import math
from functools import reduce
from itertools import count
from typing import Iterator, Iterable

from nmigen import *
from nmigen.build import ResourceError

__all__ = ["iterator_with_if_elif", "assert_is_pot", "log2", "nMin", "nMax", "nAny", "nAll",
           "max_error_freq", "delay_by", "ends_with", "connect_leds", "with_reset", "nAvrg",
           "nAbsDifference"]


def iterator_with_if_elif(iterator: Iterable, module: Module) -> Iterator:
    """
    A helper to build a priority encoder using If / Elif constructs
    :param iterator: the iterator contaianing all the elements
    :param module: the module from which m.If and m.Elif are sourced
    """
    for i, elem in enumerate(iterator):
        yield (
            module.If if i == 0 else module.Elif,
            elem
        )


def assert_is_pot(x):
    assert int(math.log2(x)) == math.log2(x), "{} is not a power of two".format(x)


def log2(x):
    assert_is_pot(x)
    return int(math.log2(x))


def nMin(a, b):
    return Mux(a < b, a, b)


def nMax(a, b):
    return Mux(a > b, a, b)


def nAny(seq):
    assert isinstance(seq, Iterable)
    return reduce(lambda a, b: a | b, seq)


def nAll(seq):
    assert isinstance(seq, Iterable)
    return reduce(lambda a, b: a & b, seq)


def max_error_freq(real_freq, requested_freq, max_error_percent=1):
    freq_error = abs((1 - (real_freq / requested_freq)) * 100)
    if freq_error > max_error_percent:
        raise ValueError("the reqested freqency {}MHz cant be synthesized by the fclk with satisfying precision ("
                         "{}% error reqested; {}% error met) the real frequency would be {}MHz"
                         .format(requested_freq / 1e6, max_error_percent, freq_error,
                                 real_freq / 1e6))
    return freq_error


def delay_by(signal, cycles, m):
    delayed_signal = signal
    for i in range(cycles):
        last = delayed_signal
        if hasattr(signal, 'name'):
            name = "{}_delayed_{}".format(signal.name, i + 1)
        else:
            name = "expression_delayed_{}".format(i + 1)
        delayed_signal = Signal.like(signal, name=name)
        m.d.sync += delayed_signal.eq(last)
    return delayed_signal


def ends_with(signal, *patterns):
    return signal.matches(
        *((("-" * (len(signal) - len(pattern))) + pattern) for pattern in patterns)
    )


def connect_leds(m, platform, signal, upper_bits=True):
    for i in count(start=0, step=1):
        try:
            m.d.comb += platform.request("led", i).eq(signal[i if not upper_bits else -(i + 1)])
        except ResourceError:
            break


def with_reset(m, signal, exclusive=False):
    domain_name = "with_reset_{}".format(signal.name)
    m.domains += ClockDomain(domain_name)
    m.d.comb += ClockSignal(domain_name).eq(ClockSignal())
    if not exclusive:
        m.d.comb += ResetSignal(domain_name).eq(ResetSignal() | signal)
    else:
        m.d.comb += ResetSignal(domain_name).eq(signal)
    return DomainRenamer(domain_name)


def nAvrg(*values):
    return sum(values) // len(values)


def nAbsDifference(a, b):
    return Mux(a > b, a - b, b - a)



from nmigen import *

"""
nMigen port of the following VHDL code, retrieved from
https://github.com/daveshah1/prjtrellis-dvi/blob/master/hdl/tmds_encoder.vhd
Original source seems to be unavailable.
Updated to match implementation from
https://www.digikey.com/eewiki/pages/viewpage.action?pageId=36569119#TMDSEncoder%28VHDL%29-AdditionalInformation
Copyright (C) 2020 Konrad Beckmann
Copyright (C) 2012 Mike Field <hamster@snap.net.nz>
License from the original source:
----------------------------------------------------------------------------------
-- Engineer: Mike Field <hamster@snap.net.nz>
-- 
-- Description: TMDS Encoder 
--     8 bits colour, 2 control bits and one blanking bits in
--       10 bits of TMDS encoded data out
--     Clocked at the pixel clock
--
----------------------------------------------------------------------------------
-- See: http://hamsterworks.co.nz/mediawiki/index.php/Dvid_test
--      http://hamsterworks.co.nz/mediawiki/index.php/FPGA_Projects
--
-- Copyright (c) 2012 Mike Field <hamster@snap.net.nz>
--
-- Permission is hereby granted, free of charge, to any person obtaining a copy
-- of this software and associated documentation files (the "Software"), to deal
-- in the Software without restriction, including without limitation the rights
-- to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
-- copies of the Software, and to permit persons to whom the Software is
-- furnished to do so, subject to the following conditions:
--
-- The above copyright notice and this permission notice shall be included in
-- all copies or substantial portions of the Software.
--
-- THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
-- IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
-- FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
-- AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
-- LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
-- OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
-- THE SOFTWARE.
--
"""


class TmdsEncoder(Elaboratable):
    def __init__(self, data, c, active):
        assert(data.shape().width == 8)
        assert(c.shape().width == 2)
        assert(active.shape().width == 1)

        self.data = data
        self.c = c
        self.blank = ~active
        self.out = Signal(10)

    def elaborate(self, platform):
        m = Module()

        data = self.data

        xored = Signal(9)
        m.d.comb += xored.eq(Cat(
            data[0],
            data[1] ^ xored[0],
            data[2] ^ xored[1],
            data[3] ^ xored[2],
            data[4] ^ xored[3],
            data[5] ^ xored[4],
            data[6] ^ xored[5],
            data[7] ^ xored[6],
            1
        ))

        xnored = Signal(9)
        m.d.comb += xnored.eq(Cat(
            data[0],
            ~(data[1] ^ xnored[0]),
            ~(data[2] ^ xnored[1]),
            ~(data[3] ^ xnored[2]),
            ~(data[4] ^ xnored[3]),
            ~(data[5] ^ xnored[4]),
            ~(data[6] ^ xnored[5]),
            ~(data[7] ^ xnored[6]),
            0
        ))

        ones = Signal(4)
        m.d.comb += ones.eq(sum(data))

        data_word = Signal(9)
        data_word_inv = Signal(9)
        disparity = Signal(Shape(4, signed=True))

        with m.If((ones > 4) | ((ones == 4) & (data[0] == 0))):
            m.d.sync += data_word.eq(xnored)
            m.d.sync += data_word_inv.eq(~xnored)
        with m.Else():
            m.d.sync += data_word.eq(xored)
            m.d.sync += data_word_inv.eq(~xored)

        # diff_q_m holds the difference between the number of ones and zeros
        diff_q_m = Signal(Shape(4, signed=True))
        data_word_ones = Signal(4)
        m.d.comb += data_word_ones.eq(sum(data_word[:8]))
        m.d.comb += diff_q_m.eq(sum(data_word[:8]) - sum(~data_word[:8]))

        with m.If(self.blank):
            with m.Switch(self.c):
                with m.Case(0b00):
                    m.d.sync += self.out.eq(0b1101010100)
                with m.Case(0b01):
                    m.d.sync += self.out.eq(0b0010101011)
                with m.Case(0b10):
                    m.d.sync += self.out.eq(0b0101010100)
                with m.Case(0b11):
                    m.d.sync += self.out.eq(0b1010101011)
            m.d.sync += disparity.eq(0)

        with m.Else():
            with m.If((disparity == 0) | (data_word_ones == 4)):
                with m.If(data_word[8] == 0):
                    m.d.sync += self.out.eq(Cat(data_word_inv[:8], data_word[8], ~data_word[8]))
                    m.d.sync += disparity.eq(disparity - diff_q_m)
                with m.Else():
                    m.d.sync += self.out.eq(Cat(data_word[:9], ~data_word[8]))
                    m.d.sync += disparity.eq(disparity + diff_q_m)
            with m.Else():
                with m.If(((disparity > 0) & (data_word_ones > 4)) |
                            ((disparity < 0) & (data_word_ones < 4))):
                    m.d.sync += self.out.eq(Cat(data_word_inv[:8], data_word[8], 0b1))
                    with m.If(data_word[8] == 0):
                        m.d.sync += disparity.eq(disparity - diff_q_m)
                    with m.Else():
                        m.d.sync += disparity.eq(disparity - diff_q_m + 2)
                with m.Else():
                    m.d.sync += self.out.eq(Cat(data_word[:9], 0b0))
                    with m.If(data_word[8] == 0):
                        m.d.sync += disparity.eq(disparity + diff_q_m - 2)
                    with m.Else():
                        m.d.sync += disparity.eq(disparity + diff_q_m)

        return m


class Decoder(Elaboratable):
    def __init__(self, data_in, data_out, c, active_data):
        assert(data_in.shape().width == 10)
        assert(data_out.shape().width == 8)
        assert(c.shape().width == 2)
        assert(active_data.shape().width == 1)

        self.data_in = data_in
        self.data_out = data_out
        self.c = c
        self.active_data = active_data

    def elaborate(self, platform):
        m = Module()

        data_in = self.data_in
        data_out = self.data_out
        c = self.c
        active_data = self.active_data

        data_delayed = Signal(10)
        sometimes_inverted = Signal(9)
        next_c = Signal(2)
        next_active_data = Signal()

        m.d.sync += c.eq(next_c)
        m.d.sync += active_data.eq(next_active_data)

        with m.If(next_active_data == 0):
            m.d.sync += data_out.eq(0)
        with m.Elif(sometimes_inverted[8] == 1):
            m.d.sync += data_out.eq(Cat(
                sometimes_inverted[0],
                sometimes_inverted[1] ^ sometimes_inverted[0],
                sometimes_inverted[2] ^ sometimes_inverted[1],
                sometimes_inverted[3] ^ sometimes_inverted[2],
                sometimes_inverted[4] ^ sometimes_inverted[3],
                sometimes_inverted[5] ^ sometimes_inverted[4],
                sometimes_inverted[6] ^ sometimes_inverted[5],
                sometimes_inverted[7] ^ sometimes_inverted[6],
            ))
        with m.Else():
            m.d.sync += data_out.eq(Cat(
                sometimes_inverted[0],
                ~(sometimes_inverted[1] ^ sometimes_inverted[0]),
                ~(sometimes_inverted[2] ^ sometimes_inverted[1]),
                ~(sometimes_inverted[3] ^ sometimes_inverted[2]),
                ~(sometimes_inverted[4] ^ sometimes_inverted[3]),
                ~(sometimes_inverted[5] ^ sometimes_inverted[4]),
                ~(sometimes_inverted[6] ^ sometimes_inverted[5]),
                ~(sometimes_inverted[7] ^ sometimes_inverted[6]),
            ))

        with m.If(data_delayed[9] == 1):
            m.d.sync += sometimes_inverted.eq(Cat(~data_delayed[:8], data_delayed[8]))
        with m.Else():
            m.d.sync += sometimes_inverted.eq(data_delayed[:9])

        with m.Switch(data_in):
            with m.Case(0b0010101011):
                m.d.sync += next_c.eq(0b01)
                m.d.sync += next_active_data.eq(0)
            with m.Case(0b1101010100):
                m.d.sync += next_c.eq(0b00)
                m.d.sync += next_active_data.eq(0)
            with m.Case(0b0101010100):
                m.d.sync += next_c.eq(0b10)
                m.d.sync += next_active_data.eq(0)
            with m.Case(0b1010101011):
                m.d.sync += next_c.eq(0b00)
                m.d.sync += next_active_data.eq(0)
            with m.Default():
                m.d.sync += next_c.eq(0b00)
                m.d.sync += next_active_data.eq(1)

        m.d.sync += data_delayed.eq(data_in)

        return m



from math import ceil
from nmigen import *
from nap import BasicStream, PacketizedStream, ControlSignal, StatusSignal

__all__ = ['Ft60xLegalizer']


class Ft60xLegalizer(Elaboratable):
    def __init__(self, input: PacketizedStream, packet_len):
        self.input = input
        self.output = BasicStream(input.payload.shape())

        # we calculate everything in bytes to make it easier to reason about
        buffer_size = 2048 * 4
        blanking = buffer_size
        aligned_len = ceil((packet_len + blanking) / buffer_size) * buffer_size
        print("ft60x paddnig:", (aligned_len - packet_len), (aligned_len - packet_len) // 4)

        self.padding = ControlSignal(16, reset=(aligned_len - packet_len) // 4)
        self.frame_len = StatusSignal(32)
        self.frame_len_changed = StatusSignal(32)

    def elaborate(self, platform):
        m = Module()

        padding_ctr = Signal.like(self.padding)
        frame_len_ctr = Signal.like(self.frame_len)

        input_transaction = self.input.ready & self.input.valid
        with m.FSM():
            with m.State("ACTIVE"):
                m.d.comb += self.input.ready.eq(self.output.ready)
                m.d.comb += self.output.valid.eq(self.input.valid)
                with m.If(self.input.payload == 0):  # we disallow the transfer of 0 to ease the alignment detection TODO: this only really works for 8 bit this way
                    m.d.comb += self.output.payload.eq(1)
                with m.Else():
                    m.d.comb += self.output.payload.eq(self.input.payload)
                with m.If(input_transaction):
                    m.d.sync += frame_len_ctr.eq(frame_len_ctr + 1)
                with m.If(self.input.last & input_transaction):
                    m.next = "PADDING"
                    m.d.sync += padding_ctr.eq(0)
            with m.State("PADDING"):
                m.d.comb += self.output.valid.eq(1)
                m.d.comb += self.input.ready.eq(0)
                m.d.comb += self.output.payload.eq(0)
                with m.If(self.output.ready):
                    with m.If(padding_ctr < self.padding - 1):
                        m.d.sync += padding_ctr.eq(padding_ctr + 1)
                        m.d.sync += frame_len_ctr.eq(frame_len_ctr + 1)
                    with m.Else():
                        m.next = "ACTIVE"
                        m.d.sync += self.frame_len.eq(frame_len_ctr + 1)
                        with m.If(self.frame_len != frame_len_ctr + 1):
                            m.d.sync += self.frame_len_changed.eq(self.frame_len_changed + 1)
                        m.d.sync += frame_len_ctr.eq(0)
        return m



from nmigen import *
from nap import BasicStream, stream_transformer

__all__ = ["Repack12BitStream"]


class Repack12BitStream(Elaboratable):
    """Repacks a packed 12 bit little endian stream to a packed 12 bit big endian stream
    This core is probably not what you want unless you want to hand of your data to a computer.
    Then do this as your very last step because everything beyond this is pure confusion.
    """
    def __init__(self, input: BasicStream, inverse=False):
        self.input = input
        assert len(self.input.payload) % 24 == 0
        self.inverse = inverse

        self.output = self.input.clone()

    def elaborate(self, platform):
        m = Module()

        stream_transformer(self.input, self.output, m, latency=0)
        for i in range(len(self.input.payload) // 24):
            input_slice = self.input.payload[i * 24:(i + 1) * 24]
            output_slice = self.output.payload[i * 24:(i + 1) * 24]

            # (assuming the least significant bit is written right and the byte with the lowest address is the rightmost byte)
            # input:  NMLKJIHG FEDCBA98 76543210
            # output: JIHGFEDC 3210NMLK BA987654

            if not self.inverse:
                m.d.comb += output_slice[0:4].eq(input_slice[4:8])
                m.d.comb += output_slice[4:8].eq(input_slice[8:12])
                m.d.comb += output_slice[8:12].eq(input_slice[20:24])
                m.d.comb += output_slice[12:16].eq(input_slice[0:4])
                m.d.comb += output_slice[16:20].eq(input_slice[12:16])
                m.d.comb += output_slice[20:24].eq(input_slice[16:20])
            else:
                m.d.comb += output_slice[4:8].eq(input_slice[0:4])
                m.d.comb += output_slice[8:12].eq(input_slice[4:8])
                m.d.comb += output_slice[20:24].eq(input_slice[8:12])
                m.d.comb += output_slice[0:4].eq(input_slice[12:16])
                m.d.comb += output_slice[12:16].eq(input_slice[16:20])
                m.d.comb += output_slice[16:20].eq(input_slice[20:24])

        return m



from nmigen import *
from ..platform_agnostic_elaboratable import PlatformAgnosticElaboratable

__all__ = ["JTAG"]


class JTAG(PlatformAgnosticElaboratable):
    def __init__(self, jtag_domain="jtag"):
        self.shift_tdi = Signal()
        self.shift_tdo = Signal()
        self.tdi = Signal()
        self.tdo = Signal()
        self.jtag_domain = jtag_domain


from nmigen import *
from nmigen.build import Clock
from nap import SocPlatform, StatusSignal
from nap.cores import DrpBridge, DrpInterface
from nap.util import decimal_range
from ..instance_helper import InstanceHelper

__all__ = ["ClockDivider", "BufIO", "BufG", "Pll", "Mmcm"]


class ClockDivider(Elaboratable):
    def __init__(self, i, divider):
        self.divider = divider
        self.i = i
        self.o = Signal()

    def elaborate(self, platform):
        return Instance(
            "BUFR",
            p_BUFR_DIVIDE=self.divider,
            p_SIM_DEVICE="7SERIES",
            i_CE=1,
            i_CLR=ResetSignal(),
            i_I=self.i,
            o_O=self.o
        )


class BufIO(Elaboratable):
    def __init__(self, i):
        self.i = i
        self.o = Signal()

    def elaborate(self, platform):
        return Instance("BUFIO", i_I=self.i, o_O=self.o)


class BufG(Elaboratable):
    def __init__(self, i):
        self.i = i
        self.o = Signal()

    def elaborate(self, platform):
        return Instance("BUFG", i_I=self.i, o_O=self.o)


_Pll = InstanceHelper("+/xilinx/cells_xtra.v", "PLLE2_ADV")


class Pll(Elaboratable):
    vco_multipliers = list(range(2, 64))
    vco_dividers = list(range(1, 56))
    output_dividers = list(range(1, 128))

    @staticmethod
    def is_valid_vco_conf(input_freq, mul, div, exception=False):
        if not mul in Pll.vco_multipliers:
            if exception:
                raise ValueError(f'mul {mul} is not an allowed multiplier')
            return False
        if not div in Pll.vco_dividers:
            if exception:
                raise ValueError(f'div {div} is not an allowed divider')
            return False
        vco_freq = input_freq * mul / div
        if 800e6 > vco_freq:
            if exception:
                raise ValueError(f'{vco_freq} is a to low vco freq. minimum is 800Mhz')
            return False
        if 1600e6 < vco_freq:
            if exception:
                raise ValueError(f'{vco_freq} is a to high vco freq. maximum is 1600Mhz')
            return False
        return True

    def __init__(self, input_freq, vco_mul, vco_div, input_domain="sync"):
        Pll.is_valid_vco_conf(input_freq, vco_mul, vco_div, exception=True)
        self._pll = _Pll(
            clkin1_period=1 / input_freq * 1e9,
            clkfbout_mult=vco_mul, divclk_divide=vco_div,
        )
        m = self.m = Module()
        m.d.comb += self._pll.clk.fbin.eq(self._pll.clk.fbout)
        m.d.comb += self._pll.clk.in_[1].eq(ClockSignal(input_domain))
        m.d.comb += self._pll.clk.insel.eq(1)  # HIGH for clkin1

        self.locked = StatusSignal()

        self._input_domain = input_domain
        self._input_freq = input_freq
        self._vco = Clock(input_freq * vco_mul / vco_div)
        self._clock_constraints = {}

    def output_domain(self, domain_name, divisor, number=None, bufg=True):
        if number is None:
            number = next(x for x in range(6) if x not in self._clock_constraints.keys())
        assert number not in self._clock_constraints.keys(), "port {} is already taken".format(number)

        assert divisor in Mmcm.output_dividers

        self._pll.parameters["CLKOUT{}_DIVIDE".format(number)] = divisor
        self._pll.parameters["CLKOUT{}_PHASE".format(number)] = 0.0

        m = self.m

        clock_signal = Signal(name="pll_out_{}".format(number), attrs={"KEEP": "TRUE"})
        m.d.comb += clock_signal.eq(self._pll.clk.out[number])

        if bufg:
            # TODO: seems to not change anything
            bufg = m.submodules["bufg_{}".format(number)] = BufG(clock_signal)
            output = bufg.o
        else:
            output = clock_signal

        m.domains += ClockDomain(domain_name)
        m.d.comb += ClockSignal(domain_name).eq(output)
        m.d.comb += ResetSignal(domain_name).eq(~self.locked)

        frequency = self._vco.frequency / divisor
        self._clock_constraints[number] = (clock_signal, frequency)
        print("PLL: creating domain '{}' with frequency {}Mhz".format(domain_name, frequency / 1e6))
        return Clock(frequency)

    def elaborate(self, platform):
        m = Module()

        m.submodules.pll_block = self._pll
        m.submodules.connections = self.m

        for i, (clock_signal, frequency) in self._clock_constraints.items():
            platform.add_clock_constraint(clock_signal, frequency)

        m.d.comb += self.locked.eq(self._pll.locked)
        m.d.comb += self._pll.rst.eq(ResetSignal(self._input_domain))

        if isinstance(platform, SocPlatform):
            m.submodules.drp_bridge = DrpBridge(DrpInterface(
                self._pll.dwe, self._pll.den, self._pll.daddr, self._pll.di, self._pll.do, self._pll.drdy, self._pll.dclk
            ))

        return m


class Mmcm(Elaboratable):
    vco_multipliers = list(decimal_range(2, 64, 0.125))
    vco_dividers = list(range(1, 106))
    output_0_dividers = list(decimal_range(1, 128, 0.125))
    output_dividers = list(range(1, 128))

    @staticmethod
    def is_valid_vco_conf(input_freq, mul, div, exception=False):
        if not mul in Mmcm.vco_multipliers:
            if exception:
                raise ValueError
            return False
        if not div in Mmcm.vco_dividers:
            if exception:
                raise ValueError
            return False
        vco_freq = input_freq * mul / div
        if 600e6 > vco_freq:
            if exception:
                raise ValueError
            return False
        if 1200e6 < vco_freq:
            if exception:
                raise ValueError
            return False
        return True

    def __init__(self, input_clock, vco_mul, vco_div, input_domain="sync"):
        Mmcm.is_valid_vco_conf(input_clock, vco_mul, vco_div, exception=True)
        self._mmcm = InstanceHelper("+/xilinx/cells_xtra.v", "MMCME2_ADV")(
            clkin1_period=1 / input_clock * 1e9,
            clkfbout_mult_f=vco_mul, divclk_divide=vco_div,
        )
        m = self.m = Module()
        m.d.comb += self._mmcm.clk.fbin.eq(self._mmcm.clk.fbout)
        m.d.comb += self._mmcm.clk.in_[1].eq(ClockSignal(input_domain))
        m.d.comb += self._mmcm.clk.ins.el.eq(1)  # HIGH for clkin1

        self.locked = StatusSignal()

        self._input_domain = input_domain
        self._input_clock = input_clock
        self._vco = Clock(input_clock * vco_mul / vco_div)
        self._clock_constraints = {}

    def output_domain(self, domain_name, divisor, number=None, bufg=True):
        if number is None:
            number = next(x for x in range(7) if x not in self._clock_constraints.keys())
        assert number not in self._clock_constraints.keys(), "port {} is already taken".format(number)

        assert divisor in (Mmcm.output_dividers if number != 0 else Mmcm.output_0_dividers)

        divide_param = "CLKOUT{}_DIVIDE{}".format(number, "_f" if number == 0 else "")
        self._mmcm.parameters[divide_param] = divisor
        self._mmcm.parameters["CLKOUT{}_PHASE".format(number)] = 0.0

        m = self.m

        clock_signal = Signal(name="mmcm_out_{}".format(number), attrs={"KEEP": "TRUE"})
        m.d.comb += clock_signal.eq(self._mmcm.clk.out[number])

        if bufg:
            # TODO: seems to not change anything
            bufg = m.submodules["bufg_{}".format(number)] = BufG(clock_signal)
            output = bufg.o
        else:
            output = clock_signal

        m.domains += ClockDomain(domain_name)
        m.d.comb += ClockSignal(domain_name).eq(output)
        m.d.comb += ResetSignal(domain_name).eq(~self.locked)

        frequency = self._vco.frequency / divisor
        self._clock_constraints[number] = (clock_signal, frequency)
        print("MMCM: creating domain '{}' with frequency {}Mhz".format(domain_name, frequency / 1e6))
        return Clock(frequency)

    def elaborate(self, platform):
        m = Module()

        m.submodules.mmcm_block = self._mmcm
        m.submodules.connections = self.m

        for i, (clock_signal, frequency) in self._clock_constraints.items():
            platform.add_clock_constraint(clock_signal, frequency)

        m.d.comb += self.locked.eq(self._mmcm.locked)
        m.d.comb += self._mmcm.rst.eq(ResetSignal(self._input_domain))

        if isinstance(platform, SocPlatform):
            m.submodules.drp_bridge = DrpBridge(DrpInterface(
                self._mmcm.dwe, self._mmcm.den, self._mmcm.daddr, self._mmcm.di, self._mmcm.do, self._mmcm.drdy,
                self._mmcm.dclk
            ))

        return m



import unittest
from collections import defaultdict
from os.path import join, dirname
from nmigen import *
from nmigen.sim import Passive
from nap import write_to_stream, ImageStream, write_frame_to_stream, read_frame_from_stream, ImageSplitter, SimPlatform
from .wavelet import Wavelet2D, MultiStageWavelet2D
import imageio
import numpy as np


class WaveletTest(unittest.TestCase):
    def test_wavelet_2d(self):
        image = imageio.imread(join(dirname(__file__), "che_128.png"))
        h, w = image.shape
        platform = SimPlatform()
        m = Module()

        input = ImageStream(8)
        transformer = m.submodules.transformer = Wavelet2D(input, w, h)

        def write_process():
            yield from write_frame_to_stream(input, image, pause=False)
            yield Passive()
            while True:
                yield from write_to_stream(input, line_last=0, frame_last=0, payload=0)
        platform.add_process(write_process, "sync")

        def read_process():
            image = (yield from read_frame_from_stream(transformer.output, timeout=1000, pause=False))
            target_image = np.copy(image)
            for y, row in enumerate(image):
                for x, px in enumerate(row):
                    target_image[y // 2 + ((y % 2) * len(image) // 2)][x // 2 + ((x % 2) * len(row) // 2)] = px
            imageio.imsave(platform.output_filename_base + ".png", target_image)
        platform.add_process(read_process, "sync")

        platform.add_sim_clock("sync", 1000e6)
        platform.sim(m)

    def test_stream_splitter(self):
        image = imageio.imread(join(dirname(__file__), "che_128.png"))
        h, w = image.shape
        platform = SimPlatform()
        m = Module()

        input = ImageStream(8)
        transformer = m.submodules.transformer = Wavelet2D(input, w, h)
        splitter = m.submodules.splitter = ImageSplitter(transformer.output, w, h)

        def write_process():
            yield from write_frame_to_stream(input, image, pause=False)
            yield Passive()
            yield from write_frame_to_stream(input, image, pause=False)
            while True:
                yield from write_to_stream(input, line_last=0, frame_last=0, payload=0)
        platform.add_process(write_process, "sync")

        for i, stream in enumerate(splitter.outputs):
            def gen_read_process():
                i_captured = i
                stream_captured = stream
                def read_process():
                    image = (yield from read_frame_from_stream(stream_captured, timeout=1000, pause=False))
                    imageio.imsave(platform.output_filename_base + "_output_{}.png".format(i_captured), image)
                return read_process
            platform.add_process(gen_read_process(), "sync")

        platform.add_sim_clock("sync", 100e6)
        platform.sim(m)

    def check_multistage(self, n):
        image = imageio.imread(join(dirname(__file__), "che_64.png"))
        h, w = image.shape
        platform = SimPlatform()
        m = Module()


        input = ImageStream(8)
        wavelet = m.submodules.wavelet = MultiStageWavelet2D(input, w, h, stages=n)

        def write_process():
            yield from write_frame_to_stream(input, image, pause=False, timeout=10000)
            yield Passive()
            while True:
                yield from write_frame_to_stream(input, image, pause=False, timeout=10000)
        platform.add_process(write_process, "sync")

        fifo_levels = defaultdict(lambda: defaultdict(int))
        def find_maximum_fifo_level():
            def find_max_levels(wavelet, level=1):
                for i, fifo in enumerate(wavelet.fifos):
                    current_level = yield fifo.r_level
                    fifo_levels[level][i] = max(current_level, fifo_levels[level][i])
                if hasattr(wavelet, 'next_stage'):
                    yield from find_max_levels(wavelet.next_stage, level + 1)
            yield Passive()
            while True:
                yield from find_max_levels(wavelet)
                yield
        platform.add_process(find_maximum_fifo_level, "sync")

        def read_process():
            for i in range(2):
                image = (yield from read_frame_from_stream(wavelet.output, timeout=1000, pause=False))
                imageio.imsave(platform.output_filename_base + str(i) + ".png", image)
        platform.add_process(read_process, "sync")

        platform.add_sim_clock("sync", 100e6)
        platform.sim(m)
        print("fifo levels:", list(fifo_levels.items()))

    def test_multistage_1(self):
        self.check_multistage(1)

    def test_multistage_2(self):
        self.check_multistage(2)

    def test_multistage_3(self):
        self.check_multistage(3)



import inspect
from os.path import join, dirname
from pathlib import Path

from nmigen import *
from nmigen import Signal
from nmigen.hdl.ast import UserValue
from nmigen.sim import Simulator

__all__ = ["SimPlatform", "FakeResource", "TristateIo", "wait_for", "pulse", "do_nothing"]


class SimPlatform:
    def __init__(self, filename=None):
        self.command_templates = []

        self.clocks = {}
        self.is_sim = True
        self.processes = []
        self.handed_out_resources = {}

        functions = []
        test_class = None
        caller_path = ""
        stack = inspect.stack()
        for frame in stack[1:]:
            if "unittest" in frame.filename:
                if not filename:
                    filename = "__".join(reversed(functions))
                    if test_class:
                        filename = f'{test_class}__{filename}'
                break
            functions.append(frame.function)
            try:
                test_class = frame.frame.f_locals['self'].__class__.__name__
            except:
                pass
            caller_path = frame.filename
        assert isinstance(filename, str)

        target_dir = join(dirname(caller_path), ".sim_results")
        Path(target_dir).mkdir(exist_ok=True)
        self.output_filename_base = join(target_dir, filename)

    def add_file(self, name, contents):
        pass

    def request(self, name, number=0):
        string = "{}#{}".format(name, number)
        if string not in self.handed_out_resources:
            self.handed_out_resources[string] = FakeResource(string, self.handed_out_resources)
        return self.handed_out_resources[string]

    def prepare(self, top_fragment, name="top", *args, **kwargs):
        # we filter all the instances out, because they give wired behaviour; TODO: this doesnt work :(
        top_fragment: Fragment = Fragment.get(top_fragment, self)

        def filter_instances(top_fragment):
            top_fragment.subfragments = [
                (filter_instances(subfragment), name)
                for subfragment, name in top_fragment.subfragments
                if not (isinstance(subfragment, Instance) and not subfragment.type.startswith('$'))
            ]
            return top_fragment

        filter_instances(top_fragment)
        return top_fragment

    def add_process(self, generator, domain=None):
        self.processes.append((generator, domain))

    def add_sim_clock(self, domain_name, frequency):
        self.clocks[domain_name] = frequency

    def sim(self, dut, testbench=None, traces=(), engine="pysim"):
        dut = self.prepare(dut)
        self.fragment = dut
        simulator = Simulator(dut, engine=engine)
        for name, frequency in self.clocks.items():
            simulator.add_clock(1 / frequency, domain=name)

        if isinstance(testbench, tuple):
            generator, domain = testbench
            self.add_process(generator, domain)
        elif inspect.isgeneratorfunction(testbench):
            self.add_process(testbench, "sync")
        elif testbench is None:
            pass
        else:
            raise TypeError("unknown type for testbench")

        for generator, domain in self.processes:
            simulator.add_sync_process(generator, domain=domain)

        print("\nwriting vcd to '{}.vcd'".format(self.output_filename_base))
        with simulator.write_vcd("{}.vcd".format(self.output_filename_base), "{}.gtkw".format(self.output_filename_base),
                                 traces=traces):
            simulator.run()


class FakeResource(UserValue):
    def __init__(self, name, handed_out_resources):
        super().__init__()
        self.handed_out_resources = handed_out_resources
        self.name = name

    def lower(self):
        return Signal(name=self.name)

    def __getattr__(self, item):
        string = "{}.{}".format(self.name, item)
        if string not in self.handed_out_resources:
            self.handed_out_resources[string] = FakeResource(string, self.handed_out_resources)
        return self.handed_out_resources[string]

    def __getitem__(self, item):
        string = "{}.{}".format(self.name, item)
        if string not in self.handed_out_resources:
            self.handed_out_resources[string] = FakeResource(string, self.handed_out_resources)
        return self.handed_out_resources[string]


class TristateIo:
    def __init__(self, shape=None):
        self.i = Signal(shape)
        self.o = Signal(shape)
        self.oe = Signal()


def wait_for(expr, timeout=100, must_clock=True):
    if must_clock:
        yield

    i = 0
    while True:
        if timeout != -1 and i >= timeout:
            raise TimeoutError("{} did not become '1' within {} cycles".format(expr, timeout))
        i += 1

        if (yield expr):
            return
        yield


def pulse(signal, length=1, after=0):
    yield signal.eq(1)
    for _ in range(length):
        yield
    yield signal.eq(0)
    for _ in range(after):
        yield


def do_nothing(length=10):
    for i in range(length):
        yield  # we expect that nothing happens here



from collections import defaultdict
from typing import Dict

from nmigen.sim import Passive
from nap import SimPlatform, write_to_stream, read_from_stream
from .axi_endpoint import AxiEndpoint, AxiResponse, AxiBurstType

__all__ = ["axil_read", "axil_write", "answer_read_burst", "answer_write_burst", "axi_ram_sim_model"]


def axil_read(axi, addr, timeout=100):
    yield from write_to_stream(axi.read_address, payload=addr)
    value, response = (yield from read_from_stream(axi.read_data, extract=("payload", "resp"), timeout=timeout))
    assert AxiResponse.OKAY.value == response
    return value


def axil_write(axi, addr, data, timeout=100):
    yield from write_to_stream(axi.write_address, payload=addr)
    yield from write_to_stream(axi.write_data, payload=data)
    response = (yield from read_from_stream(axi.write_response, extract="resp", timeout=timeout))
    assert AxiResponse.OKAY.value == response


def answer_read_burst(axi: AxiEndpoint, memory: Dict[int, int], timeout=100):
    addr, burst_len, burst_type, beat_size_bytes = yield from read_from_stream(axi.read_address, ("payload", "burst_len", "burst_type", "beat_size_bytes"), timeout)
    assert 2 ** beat_size_bytes == axi.data_bytes
    assert burst_type == AxiBurstType.INCR.value
    print("read", addr, burst_len)

    for i in range(burst_len + 1):
        yield from write_to_stream(axi.read_data, payload=memory[addr + (i * axi.data_bytes)], resp=AxiResponse.OKAY, last=(i == burst_len), timeout=timeout)


def answer_write_burst(axi: AxiEndpoint, timeout=100):
    memory = {}
    addr, burst_len, burst_type, beat_size_bytes = yield from read_from_stream(axi.write_address, ("payload", "burst_len", "burst_type", "beat_size_bytes"), timeout)
    assert 2 ** beat_size_bytes == axi.data_bytes
    assert burst_type == AxiBurstType.INCR.value
    accepted = 0
    for i in range(burst_len + 1):
        value, last, byte_strobe = yield from read_from_stream(axi.write_data, ("payload", "last", "byte_strobe"), timeout)
        if i == burst_len:
            assert last
        else:
            pass
            assert not last
        if byte_strobe != 0:
            memory[addr + i * axi.data_bytes] = value
            accepted += 1
    write_to_stream(axi.write_response, resp=AxiResponse.OKAY)
    print("wrote", memory)

    return memory, accepted


def axi_ram_sim_model(platform: SimPlatform, domain="sync"):
    mem = defaultdict(int)

    axi_writer_port = AxiEndpoint(addr_bits=32, data_bits=64, lite=False, id_bits=12)
    axi_reader_port = AxiEndpoint(addr_bits=32, data_bits=64, lite=False, id_bits=12)

    def writer_port_process():
        yield Passive()
        while True:
            memory, accepted = yield from answer_write_burst(axi_writer_port, timeout=-1)
            mem.update(memory)
    platform.add_process(writer_port_process, domain)

    def reader_port_process():
        yield Passive()
        while True:
            yield from answer_read_burst(axi_reader_port, mem, timeout=-1)
    platform.add_process(reader_port_process, domain)

    return axi_writer_port, axi_reader_port



from nmigen import *
from . import AxiEndpoint

__all__ = ["if_none_get_zynq_hp_port"]


def if_none_get_zynq_hp_port(maybe_axi_port, m, platform) -> AxiEndpoint:
    """If `maybe_axi_port` is None, grab an AXI HP port from the zynq and return it. Otherwise returns the passed in AXI port."""
    if maybe_axi_port is not None:
        assert not maybe_axi_port.is_lite
        axi = maybe_axi_port
    else:
        clock_signal = Signal()
        m.d.comb += clock_signal.eq(ClockSignal())
        axi = platform.ps7.get_axi_hp_slave(clock_signal)
    return axi



# The PluginModuleStreamer{Sink,Source} tuple allows us to stream data from the main camera FPGA (Zynq) to some
# plugin module FPGA (i.e. the usb3 plugin module machxo2)

from nmigen import *
from nap import BasicStream, ControlSignal
from nap.cores import InflexibleSinkDebug
from nap.vendor.xilinx_s7 import DDRSerializer

__all__ = ["PluginModuleStreamerTx"]


class PluginModuleStreamerTx(Elaboratable):
    def __init__(self, plugin_resource, input: BasicStream, bitclk_domain, training_pattern=0b00010110):
        self.bitclk_domain = bitclk_domain
        self.plugin_resource = plugin_resource
        self.input = input

        self.training_pattern = ControlSignal(8, reset=training_pattern)
        self.do_training = ControlSignal(reset=1)

    def elaborate(self, platform):
        m = Module()
        
        m.d.comb += self.input.ready.eq(~self.do_training)
        m.submodules.inflexible_sink_debug = InflexibleSinkDebug(self.input)

        valid = Signal()
        m.d.comb += valid.eq(self.input.valid & ~self.do_training)
        m.submodules.lane_clock = DDRSerializer(0b00001111, self.plugin_resource.clk_word, ddr_domain=self.bitclk_domain, msb_first=True)
        m.submodules.lane_valid = DDRSerializer(Repl(valid, 8), self.plugin_resource.valid, ddr_domain=self.bitclk_domain, msb_first=True)
        for i in range(4):
            value = Signal(8)
            m.submodules["lane{}".format(i)] = DDRSerializer(value, self.plugin_resource["lane{}".format(i)], ddr_domain=self.bitclk_domain, msb_first=True)
            with m.If(valid):
                m.d.comb += value.eq(self.input.payload[0+(i*8):8+(i*8)])
            with m.Else():
                m.d.comb += value.eq(self.training_pattern)

        return m



import unittest

from nmigen import *

from nap.stream.formal_util import verify_stream_output_contract, LegalStreamSource
from nap import BasicStream


class BrokenStreamSource(Elaboratable):
    def __init__(self, mode):
        self.output = BasicStream(32)
        self.mode = mode

    def elaborate(self, platform):
        m = Module()

        if self.mode == "valid_is_ready":
            m.d.comb += self.output.valid.eq(self.output.ready)
        elif self.mode == "payload_unsteady":
            m.d.sync += self.output.payload.eq(self.output.payload + 1)
            m.d.comb += self.output.valid.eq(1)
        elif self.mode == "dont_wait_for_ready":
            m.d.sync += self.output.valid.eq(~self.output.valid)

        return m


class FormalUtilTestCase(unittest.TestCase):
    def test_catch_valid_depends_on_ready(self):
        with self.assertRaises(AssertionError):
            verify_stream_output_contract(BrokenStreamSource("valid_is_ready"))

    def test_catch_changing_payload(self):
        with self.assertRaises(AssertionError):
            verify_stream_output_contract(BrokenStreamSource("payload_unsteady"))

    def test_catch_dont_wait_for_ready(self):
        with self.assertRaises(AssertionError):
            verify_stream_output_contract(BrokenStreamSource("dont_wait_for_ready"))

    def test_legal_stream_source(self):
        verify_stream_output_contract(LegalStreamSource(BasicStream(32)))



from nmigen import *
from nap import PacketizedStream, DOWNWARDS

__all__ = ["VariableWidthStream", "BitStuffer"]


class VariableWidthStream(PacketizedStream):
    """
    A stream that can indicate that only n bits of the payload are relevant.
    """

    def __init__(self, payload_shape, name=None, reset_width=0, src_loc_at=1):
        super().__init__(payload_shape, name, src_loc_at=1 + src_loc_at)
        self.current_width = Signal(range(len(self.payload)), reset=reset_width) @ DOWNWARDS


class BitStuffer(Elaboratable):
    """stuffs bits from a VariableWidthStream into a dense Stream"""
    def __init__(self, input: VariableWidthStream, output_width):
        self.input = input
        self.output = PacketizedStream(output_width, name="bit_stuffer_output")

    def elaborate(self, platform):
        m = Module()

        output_width = len(self.output.payload)

        shift_register = Signal(len(self.input.payload) + output_width)
        input_read = (self.input.ready & self.input.valid)
        output_write = (self.output.ready & self.output.valid)
        current_bits_in_shift_register = Signal(range(len(shift_register)))
        flush = Signal()

        with m.If(input_read & ~output_write):
            with m.If(self.input.last):
                m.d.sync += flush.eq(1)
            m.d.sync += current_bits_in_shift_register.eq(current_bits_in_shift_register + self.input.current_width)
            m.d.sync += shift_register.eq((self.input.payload << current_bits_in_shift_register) | shift_register)
        with m.Elif(~input_read & output_write):
            with m.If(flush & (current_bits_in_shift_register <= output_width)):
                m.d.sync += flush.eq(0)
                m.d.sync += current_bits_in_shift_register.eq(0)
            m.d.sync += current_bits_in_shift_register.eq(current_bits_in_shift_register - output_width)
            m.d.sync += shift_register.eq(shift_register[output_width:])
        with m.Elif(input_read & output_write):
            with m.If(self.input.last):
                m.d.sync += flush.eq(1)
            m.d.sync += current_bits_in_shift_register.eq(current_bits_in_shift_register + self.input.current_width - output_width)
            m.d.sync += shift_register.eq((self.input.payload << (current_bits_in_shift_register - output_width)) | (shift_register >> output_width))

        with m.If(flush & (current_bits_in_shift_register <= output_width)):
            m.d.comb += self.output.last.eq(1)
        m.d.comb += self.output.payload.eq(shift_register[:output_width])
        m.d.comb += self.output.valid.eq((current_bits_in_shift_register >= output_width) | flush)

        # do not accept a zero lenght packet if we are already presenting a valid output as this could
        # change the last signal of an ongoing transaction.
        with m.If((self.input.current_width == 0) & (current_bits_in_shift_register >= output_width)):
            m.d.comb += self.input.ready.eq(0)
        with m.Else():
            m.d.comb += self.input.ready.eq(((len(shift_register) - current_bits_in_shift_register) >= len(self.input.payload)) & ~self.output.last)

        return m


from nmigen import *

from .s7_rx_phy import MipiLaneRxPhy


class LaneWordAligner(Elaboratable):
    """A timeout based word aligner. Issues a bitslip request to the PHY if for a specified time no valid packet (as indicated by a upper layer) was received."""

    def __init__(self, pin, ddr_domain, in_packet, timeout=100_000, timeouts_to_retrain=1_000):
        self.timeouts_to_retrain = timeouts_to_retrain
        self.timeout = timeout
        self.ddr_domain = ddr_domain
        self.pin = pin

        self.enable_train_logic = Signal(reset=1)  # input; enables or disables the train logic; this is needed in multi lane setups to be able to find a training for all lanes
        self.in_packet = in_packet  # input; indicates if the upper layer is in a valid packet
        self.output = Signal(8)  # output
        self.maybe_first_packet_byte = Signal()  # indicates, that the current output word was preceded by the start of packet preamble

    def elaborate(self, platform):
        m = Module()

        phy = m.submodules.phy = MipiLaneRxPhy(self.pin, self.ddr_domain)
        m.d.comb += self.output.eq(phy.output)

        # the spec says ‘00011101’ but we need to reverse this here because we are receiving a lsb first stream
        start_of_packet = 0b10111000
        m.d.sync += self.maybe_first_packet_byte.eq((self.output == start_of_packet) & ~self.in_packet)

        with m.If(self.enable_train_logic):
            with m.FSM():
                timeout_counter = Signal(range(self.timeout))
                with m.State("UNALIGNED"):
                    with m.If(self.in_packet):
                        m.d.sync += timeout_counter.eq(0)
                        m.next = "ALIGNED"
                    with m.Elif(timeout_counter < self.timeout):
                        m.d.sync += timeout_counter.eq(timeout_counter + 1)
                    with m.Else():
                        m.d.comb += phy.bitslip.eq(1)
                        m.d.sync += timeout_counter.eq(0)

                aligned_timeouts_counter = Signal(range(self.timeouts_to_retrain))
                with m.State("ALIGNED"):
                    with m.If(self.in_packet):
                        m.d.sync += timeout_counter.eq(0)
                    with m.Elif(timeout_counter < self.timeout):
                        m.d.sync += timeout_counter.eq(timeout_counter + 1)
                    with m.Elif(aligned_timeouts_counter < self.timeouts_to_retrain):
                        m.d.sync += aligned_timeouts_counter.eq(aligned_timeouts_counter + 1)
                    with m.Else():
                        m.next = "UNALIGNED"

        return m



from enum import Enum
from functools import reduce
from nmigen import *
from nap import packed_struct, ImageStream
from .aligner import LaneWordAligner


class ShortPacketDataType(Enum):
    FRAME_START = 0x00
    FRAME_END = 0x01
    LINE_START = 0x02
    LINE_END = 0x03


class LongPacketDataType(Enum):
    RAW6 = 0x28
    RAW7 = 0x29
    RAW8 = 0x2A
    RAW10 = 0x2B
    RAW12 = 0x2C
    RAW14 = 0x2D


@packed_struct
class DataIdentifier:
    virtual_channel_identifier = unsigned(2)
    data_type = unsigned(6)

    def is_long_packet(self):
        return self.data_type <= 0x0F


@packed_struct
class PacketHeader:
    data_id: DataIdentifier
    word_count: unsigned(16)
    ecc: unsigned(8)

    def calculate_ecc(self):
        ecc_table = [
            [],
            [],
            [10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 21, 22, 23],
            [4,  5,  6,  7,  8,  9,  16, 17, 18, 19, 20, 22, 23],
            [1,  2,  3,  7,  8,  9,  13, 14, 15, 19, 20, 21, 23],
            [0,  2,  3,  5,  6,  9,  11, 12, 15, 18, 20, 21, 22],
            [0,  1,  3,  4,  6,  8,  10, 12, 14, 17, 20, 21, 22, 23],
            [0,  1,  2,  4,  5,  7,  10, 11, 13, 16, 20, 21, 22, 23]
        ]
        return Cat(reduce(lambda a, b: a ^ b, (self.as_value()[i] for i in row)) for row in ecc_table)

    def is_packet_valid(self):
        # one could also recover bit errors using the ecc; maybe do this here
        return self.calculate_ecc() == self.ecc


class MipiCSIPacketLayer(Elaboratable):
    def __init__(self, lane_word_aligner: LaneWordAligner):
        self.lane_word_aligner = lane_word_aligner

        self.output = ImageStream(32)  # we currently only support raw8 with 4 pixels / cycle

    def elaborate(self, platform):
        m = Module()

        packet_ctr = Signal(16)
        with m.FSM():
            with m.State("IDLE"):
                with m.If(self.lane_word_aligner.maybe_first_packet_byte):
                    packet_header = PacketHeader(self.lane_word_aligner.output)
                    with m.If(packet_header.is_packet_valid()):
                        m.d.comb += self.lane_word_aligner.in_packet.eq(1)
                        with m.If(packet_header.data_id.is_long_packet()):
                            m.next = "LONG_PACKET"
                            m.d.sync += packet_ctr.eq(packet_header.word_count)

            with m.State("LONG_PACKET"):
                m.d.comb += self.lane_word_aligner.in_packet.eq(1)
                with m.If(packet_ctr > 4):  # we always process 4 words per cycle
                    m.d.sync += packet_ctr.eq(packet_ctr - 4)
                with m.Else():
                    m.d.sync += packet_ctr.eq(0)
                    m.next = "IDLE"
                    # TODO: steal crc from https://gitlab.com/harmoninstruments/harmon-instruments-open-hdl/-/blob/master/Ethernet/CRC.py
                    # think about how to extract the packet footer

        return m



from unittest import TestCase
from nmigen import *
from nap import CounterStreamSource, FT601StreamSink, SimPlatform, TristateIo


class Ft601FakeResource:
    reset = Signal()
    data = TristateIo(32)
    be = TristateIo(4)
    oe = Signal()
    read = Signal()
    write = Signal()
    siwu = Signal()
    rxf = Signal()
    txe = Signal()
    clk = Signal()


class TestFt601StreamSink(TestCase):
    def test_smoke(self):
        m = Module()

        platform = SimPlatform()
        platform.add_sim_clock("sync", 50e6)
        platform.add_sim_clock("ft601", 100e6)

        ft601 = Ft601FakeResource()
        stream_counter = m.submodules.stream_counter = CounterStreamSource(32, count_if_not_ready=True)
        m.submodules.dut = FT601StreamSink(ft601, stream_counter.output)

        def testbench():
            read = []
            for i in range(3):
                yield ft601.txe.eq(1)
                written = 0
                began = False
                while True:
                    if not began:
                        if (yield ft601.write):
                            began = True
                    if began:
                        if (yield ft601.write):
                            written += 1
                            read.append((yield ft601.data.o))
                        else:
                            yield ft601.txe.eq(0)
                            break
                        if written == 2048:
                            yield ft601.txe.eq(0)
                            break
                    yield
                yield
                assert written == 2048
                for i in range(200):
                    yield
                    assert (yield ft601.write) == 0, "write was high in idle cycle {}".format(i)

            # validate the received data
            print(read)
            last = 0
            for v in read:
                assert v == last
                last += 1

        platform.sim(m, (testbench, "ft601"))



from nmigen import *
from nmigen import Elaboratable, Module, ClockSignal, ResetSignal, Cat
from nap import ControlSignal, StatusSignal, is_signal_inverted
from ..instance_helper import InstanceHelper

__all__ = ["OSerdes10", "IDelayCtrl", "IDelay", "DDRDeserializer", "DDRSerializer"]


_OSerdes = InstanceHelper("+/xilinx/cells_xtra.v", "OSERDESE2")
_ISerdes = InstanceHelper("+/xilinx/cells_xtra.v", "ISERDESE2")
_IDelay = InstanceHelper("+/xilinx/cells_xtra.v", "IDELAYE2")
_IDelayCtrl = InstanceHelper("+/xilinx/cells_xtra.v", "IDELAYCTRL")


class OSerdes10(Elaboratable):
    def __init__(self, input: Signal, pad: Signal, domain: str, domain_5x: str):
        self.pad = pad
        self.input = input

        self.domain = domain
        self.domain_5x = domain_5x

    def elaborate(self, platform):
        m = Module()

        self.invert = ControlSignal(reset=is_signal_inverted(platform, self.pad))

        data = Signal.like(self.input)
        m.d[self.domain] += data.eq(self.input ^ Repl(self.invert, len(self.input)))

        ce = Signal()
        m.d.comb += ce.eq(~ResetSignal(self.domain))

        shift = Signal(2)

        m.submodules += Instance("OSERDESE2",
                                 p_DATA_WIDTH=10, p_TRISTATE_WIDTH=1,
                                 p_DATA_RATE_OQ="DDR", p_DATA_RATE_TQ="SDR",
                                 p_SERDES_MODE="MASTER",

                                 o_OQ=self.pad,
                                 i_OCE=ce,
                                 i_TCE=0,
                                 i_RST=ResetSignal(self.domain),
                                 i_CLK=ClockSignal(self.domain_5x), i_CLKDIV=ClockSignal(self.domain),
                                 i_D1=data[0], i_D2=data[1],
                                 i_D3=data[2], i_D4=data[3],
                                 i_D5=data[4], i_D6=data[5],
                                 i_D7=data[6], i_D8=data[7],

                                 i_SHIFTIN1=shift[0], i_SHIFTIN2=shift[1],
                                 )

        m.submodules += Instance("OSERDESE2",
                                 p_DATA_WIDTH=10, p_TRISTATE_WIDTH=1,
                                 p_DATA_RATE_OQ="DDR", p_DATA_RATE_TQ="SDR",
                                 p_SERDES_MODE="SLAVE",

                                 i_OCE=ce,
                                 i_TCE=0,
                                 i_RST=ResetSignal(self.domain),
                                 i_CLK=ClockSignal(self.domain_5x), i_CLKDIV=ClockSignal(self.domain),
                                 i_D1=0, i_D2=0,
                                 i_D3=data[8], i_D4=data[9],
                                 i_D5=0, i_D6=0,
                                 i_D7=0, i_D8=0,

                                 i_SHIFTIN1=0, i_SHIFTIN2=0,
                                 o_SHIFTOUT1=shift[0], o_SHIFTOUT2=shift[1]
                                 )

        return m


class IDelayCtrl(Elaboratable):
    def __init__(self, refclk_domain):
        self.refclk_domain = refclk_domain
        self.ready = StatusSignal()

    def elaborate(self, platform):
        m = Module()

        idelay_ctl = m.submodules.idelay_ctl = _IDelayCtrl()
        m.d.comb += self.ready.eq(idelay_ctl.rdy)
        m.d.comb += idelay_ctl.refclk.eq(ClockSignal(self.refclk_domain))
        m.d.comb += idelay_ctl.rst.eq(ResetSignal(self.refclk_domain))

        return m


class IDelay(Elaboratable):
    def __init__(self, pin):
        self.pin = pin

        self.delay = ControlSignal(5)
        self.load = ControlSignal()

        self.output = Signal()

    def elaborate(self, platform):
        m = Module()

        idelay = m.submodules.idelay = _IDelay(
            delay_src="iDataIn",
            signal_pattern="data",
            cinvctrl_sel=False,
            high_performance_mode=True,
            refclk_frequency=200.0,
            pipe_sel=False,
            idelay_type="var_load",
            idelay_value=0
        )
        m.d.comb += idelay.c.eq(ClockSignal())  # this is really the clock to which the control inputs are syncronous!
        m.d.comb += idelay.ld.eq(self.load)
        m.d.comb += idelay.ldpipeen.eq(0)
        m.d.comb += idelay.ce.eq(0)
        m.d.comb += idelay.inc.eq(0)
        m.d.comb += idelay.cntvalue.in_.eq(self.delay)
        m.d.comb += idelay.idatain.eq(self.pin)
        m.d.comb += self.output.eq(idelay.data.out)

        return m


class DDRSerializer(Elaboratable):
    def __init__(self, value, pad, ddr_domain, bit_width=8, msb_first=False):
        self.msb_first = msb_first
        self.bit_width = bit_width
        self.ddr_domain = ddr_domain
        self.value = value
        self.pad = pad

    def elaborate(self, platform):
        m = Module()

        self.invert = ControlSignal(reset=is_signal_inverted(platform, self.pad))

        oserdes = m.submodules.oserdes = _OSerdes(
            data_width=self.bit_width,
            tristate_width=1,
            data_rate_oq="ddr",
            serdes_mode="master",
            data_rate_tq="buf"
        )
        m.d.comb += oserdes.oce.eq(1)
        m.d.comb += oserdes.clk.eq(ClockSignal(self.ddr_domain))
        m.d.comb += oserdes.clkdiv.eq(ClockSignal())
        m.d.comb += oserdes.rst.eq(ResetSignal())
        m.d.comb += Cat(oserdes.d[i] for i in (range(1, 9) if self.msb_first else reversed(range(1, 9)))).eq(self.value ^ self.invert)
        m.d.comb += self.pad.eq(oserdes.oq)

        return m


class DDRDeserializer(Elaboratable):
    def __init__(self, pad, ddr_domain, bit_width=8, msb_first=False):
        self.msb_first = msb_first
        self.bit_width = bit_width
        self.ddr_domain = ddr_domain
        self.pad = pad

        self.bitslip = Signal()
        self.output = Signal(bit_width)

    def elaborate(self, platform):
        m = Module()

        iserdes = m.submodules.iserdes = _ISerdes(
            data_width=self.bit_width,
            data_rate="ddr",
            serdes_mode="master",
            interface_type="networking",
            num_ce=1,
            iobDelay="ifd",
        )
        m.d.comb += iserdes.ddly.eq(self.pad)
        m.d.comb += iserdes.ce[1].eq(1)
        m.d.comb += iserdes.clk.eq(ClockSignal(self.ddr_domain))
        m.d.comb += iserdes.clkb.eq(~ClockSignal(self.ddr_domain))
        m.d.comb += iserdes.rst.eq(ResetSignal())
        m.d.comb += iserdes.clkdiv.eq(ClockSignal())
        m.d.comb += self.output.eq(Cat(iserdes.q[i] for i in (range(1, 9) if self.msb_first else reversed(list(range(1, 9))))))
        m.d.comb += iserdes.bitslip.eq(self.bitslip)

        return m



from nmigen import *
from nap.stream import BasicStream, stream_transformer

__all__ = ["StreamBuffer"]


class StreamBuffer(Elaboratable):
    """Basically a 1 deep Stream FIFO. Can be used to improve timing or to make outputs compliant with the Stream contract"""
    def __init__(self, input: BasicStream):
        self.input = input
        self.output = input.clone()

    def elaborate(self, platform):
        m = Module()

        stream_transformer(self.input, self.output, m, latency=1)
        with m.If(self.input.ready & self.input.valid):
            m.d.sync += self.output.connect_upstream(self.input, exclude=["ready", "valid"])

        return m



from typing import Tuple
from nmigen import *
from nmigen.hdl.ast import Operator
from nap import StatusSignal
from . import ImageStream

__all__ = ["ImageConvoluter"]


class ImageConvoluter(Elaboratable):
    def __init__(self, input: ImageStream, transformer_function, width, height):
        self.input = input

        self.input_x = Signal(range(width))
        self.input_y = Signal(range(height))

        self.output_x = Signal(range(width))
        self.output_y = Signal(range(height))

        self.image_proxy = ImageProxy(self.input.payload.shape(), self.output_x, self.output_y)
        self.output_payload = transformer_function(self.output_x, self.output_y, self.image_proxy)
        self.shift_x = -min([x for x, y in self.image_proxy.offset_mapping.keys()] + [0])
        self.shift_y = -min([y for x, y in self.image_proxy.offset_mapping.keys()] + [0])

        self.width = width
        self.height = height

        self.delay_needed = self.shift_x + (width * self.shift_y)
        self.delayed_cycles = StatusSignal(range(self.delay_needed))

        self.output = ImageStream(Value.cast(self.output_payload).shape(), name="image_convoluter_output")

    def elaborate(self, platform):
        m = Module()

        read_input = self.input.ready & self.input.valid
        write_output = self.output.ready & self.output.valid

        available_pixels = {}
        if len(self.image_proxy.offset_mapping.keys()) > 0:
            needed_rows = max(y + self.shift_y for x, y in self.image_proxy.offset_mapping.keys())
            for y in range(needed_rows + 1):
                if y == 0:
                    available_pixels[(0, y)] = self.input.payload
                else:
                    memory = Memory(width=len(self.input.payload), depth=self.width)
                    write_port = memory.write_port()
                    m.submodules += write_port
                    with m.If(read_input):
                        m.d.comb += write_port.addr.eq(self.input_x)
                        m.d.comb += write_port.data.eq(available_pixels[0, y-1])
                        m.d.comb += write_port.en.eq(1)
                    read_port = memory.read_port(transparent=False)
                    m.submodules += read_port
                    m.d.comb += read_port.en.eq(1)
                    with m.If(read_input):
                        m.d.comb += read_port.addr.eq((self.input_x + 1) % self.width)
                    with m.Else():
                        m.d.comb += read_port.addr.eq(self.input_x)
                    available_pixels[(0, y)] = Signal.like(self.input.payload)
                    m.d.comb += available_pixels[(0, y)].eq(read_port.data)

                needed_pixels = max(x + self.shift_x for x, y in self.image_proxy.offset_mapping.keys() if y == y)
                for x in range(1, needed_pixels + 1):
                    available_pixels[(x, y)] = Signal.like(self.input.payload)
                    with m.If(read_input):
                        m.d.sync += available_pixels[(x, y)].eq(available_pixels[(x - 1, y)])

        with m.If(read_input):
            with m.If(~self.input.line_last):
                m.d.sync += self.input_x.eq(self.input_x + 1)
            with m.Else():
                m.d.sync += self.input_x.eq(0)
                with m.If(~self.input.frame_last):
                    m.d.sync += self.input_y.eq(self.input_y + 1)
                with m.Else():
                    m.d.sync += self.input_y.eq(0)

        with m.If(write_output):
            with m.If(self.output_x < self.width - 1):
                m.d.sync += self.output_x.eq(self.output_x + 1)
            with m.Else():
                m.d.sync += self.output_x.eq(0)
                m.d.comb += self.output.line_last.eq(1)
                with m.If(self.output_y < self.height - 1):
                    m.d.sync += self.output_y.eq(self.output_y + 1)
                with m.Else():
                    m.d.sync += self.output_y.eq(0)
                    m.d.comb += self.output.frame_last.eq(1)
        for (x, y), signal in self.image_proxy.offset_mapping.items():
            m.d.comb += signal.eq(available_pixels[x + self.shift_x, y + self.shift_y])
        m.d.comb += self.output.payload.eq(self.output_payload)
        
        with m.If(read_input & ~write_output):
            m.d.sync += self.delayed_cycles.eq(self.delayed_cycles + 1)
        with m.Elif(~read_input & write_output):
            m.d.sync += self.delayed_cycles.eq(self.delayed_cycles - 1)

        with m.If(self.delayed_cycles < self.delay_needed):
            m.d.comb += self.input.ready.eq(1)
        with m.Elif(self.delayed_cycles > self.delay_needed):
            m.d.comb += self.output.valid.eq(1)
        with m.Else():
            m.d.comb += self.output.valid.eq(self.input.valid)
            m.d.comb += self.input.ready.eq(self.output.ready)

        return m


class ImageProxy:
    def __init__(self, shape, x_signal, y_signal):
        self.y_signal = y_signal
        self.x_signal = x_signal
        self.shape = shape
        self.offset_mapping = {}

    @classmethod
    def _offset_from_expression(cls, expr, allowed_variable):
        e = ValueError(
            "the value {!r} is not allowed here. only terms that are {!r} +/- a constant are allowed"
                .format(expr, allowed_variable))
        if isinstance(expr, Signal):
            if not expr is allowed_variable:
                raise e
            return 0
        elif isinstance(expr, Operator):
            if not expr.operator in ("+", "-"):
                raise e
            if (not len(expr.operands) == 2) and (expr.operands[0] is allowed_variable) \
                    and (isinstance(expr.operands[1], Const)):
                raise e
            return expr.operands[1].value * (-1 if expr.operator == "-" else 1)
        else:
            raise e

    def __getitem__(self, item):
        assert isinstance(item, Tuple) and len(item) == 2
        offset = -self._offset_from_expression(item[0], self.x_signal), \
                 -self._offset_from_expression(item[1], self.y_signal)

        if offset not in self.offset_mapping:
            self.offset_mapping[offset] = Signal(self.shape)
        return self.offset_mapping[offset]



import re
import subprocess
from typing import Dict

from nmigen.back import rtlil


def get_module_sizes(module, *args, **kwargs):
    rtlil_text = rtlil.convert(module, *args, **kwargs)

    script = """
        read_ilang <<rtlil
        {}
        rtlil
        
        expose top
        
        synth_xilinx -abc9
    """.format(rtlil_text)

    popen = subprocess.Popen(["yosys", "-s", "-"],
                             stdin=subprocess.PIPE,
                             stdout=subprocess.PIPE,
                             stderr=subprocess.PIPE,
                             encoding="utf-8")
    output, error = popen.communicate(script)

    return {
        re.findall("== (.*?) ===", section)[0]: re.findall("Estimated number of LCs:\\W*(\\d+)", section)[0]
        for section in re.findall("== .*? ===.*?=", output, flags=re.DOTALL)
    }


def print_module_sizes(module, *args, **kwargs):
    module_sizes: Dict[str, str] = get_module_sizes(module, *args, **kwargs)
    max_module_size = max(int(v) for v in module_sizes.values())

    for name, size in module_sizes.items():
        print(name.ljust(30), size.ljust(3), "".join("=" for _ in range(int(int(size) / max_module_size * 80))))


from nmigen import *
from nap import packed_struct

__all__ = ["RGB24", "RGB565"]


def gen_rgb_n(r_bits, g_bits, b_bits):
    @packed_struct
    class RGB:
        r: unsigned(r_bits)
        g: unsigned(g_bits)
        b: unsigned(b_bits)

        def brightness(self):
            return self.r + self.g + self.b

    return RGB


RGB24 = gen_rgb_n(8, 8, 8)
RGB565 = gen_rgb_n(5, 6, 5)



from nmigen.build import Resource, Subsignal, Pins, DiffPairs, Connector, Attrs, PinsN, DiffPairsN
from nmigen_boards.zturn_lite_z010 import ZTurnLiteZ010Platform
from .plugins.plugin_connector import add_plugin_connector

__all__ = ["MicroR2Platform"]


class MicroR2Platform(ZTurnLiteZ010Platform):
    def toolchain_program(self, products, name, **kwargs):
        pass

    def __init__(self):
        super().__init__()
        self.add_resources([
            Resource("sensor", 0,
                     Subsignal("shutter", Pins("25", dir='o', conn=("expansion", 0)), Attrs(IOSTANDARD="LVCMOS25")),
                     Subsignal("trigger", Pins("27", dir='o', conn=("expansion", 0)), Attrs(IOSTANDARD="LVCMOS25")),
                     Subsignal("reset", PinsN("31", dir='o', conn=("expansion", 0)), Attrs(IOSTANDARD="LVCMOS25")),
                     Subsignal("clk", Pins("33", dir='o', conn=("expansion", 0)), Attrs(IOSTANDARD="LVCMOS25")),
                     Subsignal("lvds_clk", DiffPairs("51", "53", dir='i', conn=("expansion", 0)), Attrs(IOSTANDARD="LVDS_25", DIFF_TERM="TRUE")),
                     Subsignal("lvds", DiffPairsN("41 45 55 65", "43 47 57 67", dir='i', conn=("expansion", 0)), Attrs(IOSTANDARD="LVDS_25", DIFF_TERM="TRUE")),
                     ),
            Resource("i2c", 0,
                     Subsignal("scl", Pins("35", dir='io', conn=("expansion", 0)), Attrs(IOSTANDARD="LVCMOS25")),
                     Subsignal("sda", Pins("37", dir='io', conn=("expansion", 0)), Attrs(IOSTANDARD="LVCMOS25")),
                     ),
            Resource("ws2812", 0, Pins("56", dir='o', conn=("expansion", 0)), Attrs(IOSTANDARD="LVCMOS25")),
            Resource("encoder", 0,
                     Subsignal("quadrature", Pins("58 68", dir='i', conn=("expansion", 0))),
                     Subsignal("push", Pins("66", dir='i', conn=("expansion", 0)))
                     ),
        ])

        add_plugin_connector(
            platform=self, number="south", conn=("expansion", 0),
            lvds=["21 23", "3 5", "9 11", "13 15", "4 6", "10 <-> 12"],
            gpio=[71, 73, 63, 61, 64, 62, 75, 77],
            i2c=[46, 48],
        )
        add_plugin_connector(
            platform=self, number="north", conn=("expansion", 0),
            lvds=["14 <-> 16", "22 24", "26 28", "32 34", "36 38", "42 <-> 44"],
            gpio=[76, 78, 80, 84, 86, 88, 90, 74],
            i2c=[52, 54]
        )

        self.add_connectors([
            Connector("pmod", "north", "110 106 100 96 - - 108 104 98 94 - -", conn=("expansion", 0), ),
            Connector("pmod", "south", "97 95 89 85 - - 99 93 87 83 - -", conn=("expansion", 0), ),
            Connector("pmod", "east", "103 105 107 109 - -", conn=("expansion", 0), ),
        ])



import unittest

from . import CounterStreamSource
from nap import verify_stream_output_contract


class CounterStreamSourceTest(unittest.TestCase):
    def test_stream_contract(self):
        verify_stream_output_contract(CounterStreamSource(32))



# a decorator to mark methods in elaboratables that should end up in the pydriver
# all Drivermethod instances are collected and shipped with the pydriver

__all__ = ["driver_method", "driver_property"]

class DriverMethod:
    def __init__(self, function, is_property):
        self.is_property = is_property
        self.function = function

    def __repr__(self):
        if self.is_property:
            return "driver_property"
        else:
            return "driver_method()"


def driver_method(function):
    return DriverMethod(function, is_property=False)


def driver_property(function):
    return DriverMethod(function, is_property=True)



from nmigen.build import Resource, Pins, DiffPairs, Attrs, Subsignal
from nmigen.vendor.lattice_machxo_2_3l import LatticeMachXO2Platform
from nmigen_boards.microzed_z020 import MicroZedZ020Platform

from .plugins import add_plugin_connector

__all__ = ["BetaPlatform", "BetaRFWPlatform"]


class BetaPlatform(MicroZedZ020Platform):
    def __init__(self):
        super().__init__()
        self.connect_mainboard()

    def connect_mainboard(self):
        add_plugin_connector(
            platform=self, number="south", conn=("JX2", 0),
            lvds=["94 96", "93 95", "97 99", "87 89", "81 83", "73 75"],
        )
        add_plugin_connector(
            platform=self, number="north", conn=("JX1", 0),
            lvds=["68 70", "74 76", "82 84", "92 94", "93 91", "89 87"]
        )

        # TODO: add ext & shield connectors (but how?)
        #       best would be a way to (transpranetly) handle the routing fabrics
        self.add_resources([
            Resource("routing", 'east', DiffPairs('29', '31', dir='io', conn=("JX2", 0)), Attrs(IOSTANDARD="LVCMOS33")),
            Resource("routing", 'west', Pins("56", dir='o', conn=("JX1", 0)), Attrs(IOSTANDARD="LVCMOS33")),
        ])


class BetaRFWPlatform(LatticeMachXO2Platform):
    device = "LCMXO2-2000HC"
    package = "TG100"
    speed = "6"

    resources = [
        Resource(
            "pic_io", 0,
            Subsignal("ss", Pins("27", dir="io"), Attrs(IO_TYPE="LVCMOS33", PULLMODE="UP", DRIVE="4")),
            Subsignal("sck", Pins("31", dir="io"), Attrs(IO_TYPE="LVCMOS33", PULLMODE="UP", DRIVE="4")),
            Subsignal("sdo", Pins("32", dir="io"), Attrs(IO_TYPE="LVCMOS33", PULLMODE="UP", DRIVE="4")),
            Subsignal("pb22b", Pins("47", dir="io"), Attrs(IO_TYPE="LVCMOS33", PULLMODE="UP", DRIVE="4")),
            Subsignal("sn", Pins("48", dir="io"), Attrs(IO_TYPE="LVCMOS33", PULLMODE="UP", DRIVE="4")),
            Subsignal("sdi", Pins("49", dir="io"), Attrs(IO_TYPE="LVCMOS33", PULLMODE="UP", DRIVE="4")),
            Subsignal("done", Pins("76", dir="io"), Attrs(IO_TYPE="LVCMOS33", PULLMODE="UP", DRIVE="4")),
            Subsignal("initn", Pins("77", dir="io"), Attrs(IO_TYPE="LVCMOS33", PULLMODE="UP", DRIVE="4")),
        )
    ]
    connectors = []

    def __init__(self):
        super().__init__()

        add_plugin_connector(self, "north", gpio=["83", "78", "74", "75", "71", "70", "69", "68"])
        add_plugin_connector(self, "south", gpio=["1", "2", "98", "99", "96", "97", "84", "87"])



from nmigen import *
from nap import PacketizedStream, ImageStream, VariableWidthStream
from nap.cores import BitStuffer, HuffmanEncoder, ZeroRleEncoder, ImageStream2PacketizedStream
from .wavelet import MultiStageWavelet2D


class WaveletCompressor(Elaboratable):
    """compresses an image stream using a wavelet compression algorithm"""
    def __init__(self, input: ImageStream, width, height):
        self.input = input
        self.width = width
        self.height = height

        max_input_word = 2**len(self.input.payload)
        self.possible_run_lengths = [2, 4, 8, 16, 32, 64, 128, 256]
        self.huffmann_frequencies = {
            **{k: 1 for k in range(max_input_word)},
            **{k: 10 for k in range(max_input_word, max_input_word + len(self.possible_run_lengths))}
        }

        self.output = PacketizedStream(self.input.payload.shape())

    def elaborate(self, platform):
        m = Module()

        wavelet = m.submodules.wavelet = MultiStageWavelet2D(self.input, self.width, self.height, stages=3)
        packetizer = m.submodules.packetizer = ImageStream2PacketizedStream(wavelet.output)

        bit_stuffing_input = VariableWidthStream(self.input.payload.shape(), reset_width=len(self.input.payload))
        with m.If(packetizer.output.is_hf):
            rle_input = PacketizedStream()
            m.d.comb += rle_input.connect_upstream(packetizer.output)
            rle = m.submodules.rle = ZeroRleEncoder(rle_input, self.possible_run_lengths)
            huffman = m.submodules.huffman = HuffmanEncoder(rle.output)
            m.d.comb += bit_stuffing_input.connect_upstream(huffman.output)
        with m.Else():
            m.d.comb += bit_stuffing_input.connect_upstream(packetizer.output)

        bit_stuffing = m.submodules.bit_stuffing = BitStuffer(bit_stuffing_input, len(self.output.payload))
        m.d.comb += self.output.connect_upstream(bit_stuffing.output)

        return m



from nmigen import *
from nap import BasicStream, DOWNWARDS
from . import StreamInfo

__all__ = ["StreamResizer", "StreamGearbox", "SimpleStreamGearbox"]


class StreamResizer(Elaboratable):
    """Simply resizing a Stream by truncating or zero extending the payload"""

    def __init__(self, input: BasicStream, target_width, upper_bits=False):
        self.input = input
        self.output = input.clone(name="resizer_output")
        self.output.payload = Signal(target_width) @ DOWNWARDS
        self.upper_bits = upper_bits

    def elaborate(self, platform):
        m = Module()
        m.d.comb += self.output.connect_upstream(self.input, exclude=["payload"])
        if not self.upper_bits:
            m.d.comb += self.output.payload.eq(self.input.payload)
        else:
            m.d.comb += self.output.payload.eq(self.input.payload[-len(self.output.payload):])
        return m


class StreamGearbox(Elaboratable):
    """Resize a Stream by 'Gearing' it up / down (changing the word rate)"""
    # TODO: add flushing on first & last
    # (maybe add valid logic wit sub word granularity?)
    # do we need streams with more granular enables (also for more than 1 px / cycle throughput blocks?)

    def __init__(self, input: BasicStream, target_width):
        self.input = input
        self.output = input.clone(name="gearbox_output")
        self.output.payload = Signal(target_width) @ DOWNWARDS

    def elaborate(self, platform):
        m = Module()

        input_width = len(self.input.payload)
        output_width = len(self.output.payload)

        shift_register = Signal(input_width + output_width)
        input_read = (self.input.ready & self.input.valid)
        output_write = (self.output.ready & self.output.valid)
        current_bits_in_shift_register = Signal(range(len(shift_register)))

        for k, v in self.input.out_of_band_signals.items():
            if not (k == "payload" or (k.endswith("last") and len(v) == 1) or (k.endswith("first") and len(v) == 1)):
                raise ValueError("payload signal {} of input has unknown role. dont know what do do with it")

        last_first_shift_registers = {k: Signal.like(shift_register, name="{}_shift_register".format(k))
                                for k, v in self.input.out_of_band_signals.items() if (k.endswith("last") or k.endswith("first"))}

        shift_registers = [(shift_register, self.input.payload)]
        shift_registers += [(reg, (self.input[k] << ((input_width - 1) if k.endswith("last") else 0)))
                            for k, reg in last_first_shift_registers.items()]

        with m.If(input_read & ~output_write):
            m.d.sync += current_bits_in_shift_register.eq(current_bits_in_shift_register + input_width)
            for reg, payload in shift_registers:
                m.d.sync += reg.eq((payload << current_bits_in_shift_register) | reg)
        with m.Elif(~input_read & output_write):
            m.d.sync += current_bits_in_shift_register.eq(current_bits_in_shift_register - output_width)
            for reg, payload in shift_registers:
                m.d.sync += reg.eq(reg[output_width:])
        with m.Elif(input_read & output_write):
            m.d.sync += current_bits_in_shift_register.eq(current_bits_in_shift_register + input_width - output_width)
            for reg, payload in shift_registers:
                m.d.sync += reg.eq((payload << (current_bits_in_shift_register - output_width)) | (reg >> output_width))

        m.d.comb += self.output.payload.eq(shift_register[:output_width])
        m.d.comb += self.output.valid.eq(current_bits_in_shift_register >= output_width)
        m.d.comb += self.input.ready.eq((len(shift_register) - current_bits_in_shift_register) >= input_width)

        m.d.comb += [self.output[k].eq(reg[:output_width] != 0) for k, reg in last_first_shift_registers.items()]

        return m


class SimpleStreamGearbox(Elaboratable):
    def __init__(self, input: BasicStream, target_width):
        self.input = input
        self.output = input.clone(name="gearbox_output")
        self.output.payload = Signal(target_width) @ DOWNWARDS

        self.input_width = len(self.input.payload)
        self.output_width = target_width
        self.division_factor = self.input_width / target_width

        assert target_width < self.input_width
        assert self.division_factor % 1 == 0
        self.division_factor = int(self.division_factor)

    def elaborate(self, platform):
        m = Module()

        m.submodules.input_stream_info = StreamInfo(self.input)
        m.submodules.output_stream_info = StreamInfo(self.output)

        input_read = (self.input.ready & self.input.valid)
        output_write = (self.output.ready & self.output.valid)

        for k, v in self.input.out_of_band_signals.items():
            if not (k == "payload" or (k.endswith("last") and len(v) == 1)):
                raise ValueError("payload signal {} of input has unknown role. dont know what do do with it")

        last_signals = {
            k: Signal(1, name="{}_store".format(k))
            for k, v in self.input.out_of_band_signals.items()
            if k.endswith("last")
        }

        reg = Signal(self.input_width - self.output_width)
        state = Signal(range(self.division_factor))
        with m.If(state == 0):
            m.d.comb += self.input.ready.eq(self.output.ready)
            m.d.comb += self.output.valid.eq(self.input.valid)
            m.d.comb += self.output.payload.eq(self.input.payload[:self.output_width])
            with m.If(input_read):
                m.d.sync += reg.eq(self.input.payload[self.output_width:])
                m.d.sync += state.eq(state + 1)
                for k, v in last_signals.items():
                    m.d.sync += v.eq(self.input[k])
        with m.Else():
            m.d.comb += self.output.payload.eq(reg)
            m.d.comb += self.output.valid.eq(1)
            m.d.comb += self.input.ready.eq(0)
            with m.If(output_write):
                m.d.sync += reg.eq(reg[self.output_width:])
                m.d.sync += state.eq(state + 1)
            with m.If(state == self.division_factor - 1):
                for k, v in last_signals.items():
                    m.d.comb += self.output[k].eq(v)

        return m



# An experiment for debugging the JTAG interface on the usb3 plugin module by using the USB3 output
# as a usb3 output.
from nmigen import *
from nap import *
from nap.vendor.lattice_machxo2 import Osc


class Top(Elaboratable):
    def __init__(self):
        self.counter = StatusSignal(32)
        self.test_reg32 = ControlSignal(32)

    def elaborate(self, platform):
        m = Module()
        m.submodules.osc = Osc(freq=53.2e6)
        m.d.sync += self.counter.eq(self.counter + 1)
        m.d.comb += platform.request("led", 0).eq(self.counter[24])

        debug_stream_source = BasicStream(32)
        m.d.comb += debug_stream_source.valid.eq(1)
        m.d.comb += debug_stream_source.payload.eq(platform.jtag_debug_signals)

        ft601 = platform.request("ft601")
        m.submodules.ft601 = FT601StreamSink(ft601, debug_stream_source, domain_name="ft601")

        return m


if __name__ == "__main__":
    cli(Top, runs_on=(Usb3PluginPlatform,), possible_socs=(JTAGSocPlatform,))



import argparse
import inspect
import os
import pickle
import sys
from collections import OrderedDict
from datetime import datetime
from glob import glob
from hashlib import sha256
from json import dumps
from os import stat, path
from os.path import exists, isdir
from warnings import warn

from nmigen.build.run import LocalBuildProducts

__all__ = ["cli"]


def hash_build_plan(build_plan_files):
    build_plan_files = {k: v.decode("utf-8") if isinstance(v, bytes) else v for k, v in build_plan_files.items()}
    json_repr = dumps(build_plan_files)
    return sha256(json_repr.encode("utf-8")).hexdigest()


def get_previous_build_dir(basename):
    build_files = glob("build/{}*".format(basename))
    build_files = [path for path in build_files if isdir(path)]
    if not build_files:
        return None
    sorted_build_files = sorted(build_files, key=lambda x: stat(x).st_mtime, reverse=True)
    return sorted_build_files[0].replace('build/build_', '').replace('.sh', '')


def cli(top_class, runs_on, possible_socs=(None,)):
    parser = argparse.ArgumentParser()
    parser.add_argument('-e', '--elaborate', help='Elaborates the experiment', action="store_true")
    parser.add_argument('-b', '--build',
                        help='builds the experiment with the vendor tools if the elaboration result changed; implies -e',
                        action="store_true")
    parser.add_argument('-p', '--program', help='programs the board; programs the last build if used without -b',
                        action="store_true")

    platform_choices = [plat.__name__.replace("Platform", "") for plat in runs_on]
    default = platform_choices[0] if len(platform_choices) == 1 else None
    parser.add_argument('-d', '--device', help='specify the device to build for', choices=platform_choices,
                        required=default is None, default=default)

    soc_choices = [plat.__name__.replace("SocPlatform", "") if plat is not None else "None" for plat in possible_socs]
    default = soc_choices[0] if len(soc_choices) == 1 else None
    parser.add_argument('-s', '--soc', help='specifies the soc platform to build for', choices=soc_choices, default=default, required=default is None)
    parser = parser
    args = parser.parse_args()

    hardware_platform = getattr(__import__('nap.platform'), "{}Platform".format(args.device))
    if args.soc != 'None':
        soc_platform = getattr(__import__('nap.soc.platform'), "{}SocPlatform".format(args.soc))
        assert soc_platform in possible_socs
        platform = soc_platform(hardware_platform())
    else:
        assert None in possible_socs
        platform = hardware_platform()
    top_class = top_class

    name = inspect.stack()[1].filename.split("/")[-1].replace(".py", "")
    dir_basename = "{}_{}_{}".format(name, args.device, args.soc)

    if not (args.program or args.build or args.elaborate):
        print("no action specified")
        parser.print_help(sys.stderr)
        exit(-1)

    if args.elaborate or args.build:
        build_plan = platform.build(
            top_class(),
            name=name,
            do_build=False,
        )

    if args.build:
        needs_rebuild = True
        build_plan_hash = hash_build_plan(build_plan.files)
        previous_build_dir = get_previous_build_dir(dir_basename)
        if previous_build_dir:
            try:
                old_build_plan_files = OrderedDict(
                    (k, open(path.join(previous_build_dir, k)).read())
                    for k, v in build_plan.files.items()
                )
                old_build_plan_hash = hash_build_plan(old_build_plan_files)
                if old_build_plan_hash == build_plan_hash and exists(
                        path.join(previous_build_dir, 'extra_files.pickle')):
                    needs_rebuild = False
            except FileNotFoundError as e:
                warn("something went wrong while determining if a rebuild is nescessary :(. "
                     "Rebuilding unconditionally ...\n" + str(e))

        if needs_rebuild:
            build_subdir = dir_basename + datetime.now().strftime("__%d_%b_%Y__%H_%M_%S")
            build_path = path.join("build", build_subdir)
            build_plan.execute_local(build_path)
            with open(path.join(build_path, 'extra_files.pickle'), 'wb') as f:
                pickle.dump(platform.extra_files, f)

    if args.program:
        previous_build_dir = get_previous_build_dir(dir_basename)
        with open(path.join(previous_build_dir, 'extra_files.pickle'), 'rb') as f:
            platform.extra_files = pickle.load(f)
        cwd = os.getcwd()
        try:
            os.chdir(previous_build_dir)
            platform.toolchain_program(LocalBuildProducts(os.getcwd()), name=name)
        finally:
            os.chdir(cwd)



# remaps the cmv12k output lanes into an ImageStream with subsequent pixels
# the raw lane output of the cmv12k is a bit wired

from typing import List

from nmigen import *
from nap import BasicStream, packed_struct
from nap.cores import ImageStream, BufferedSyncStreamFIFO

__all__ = ["Cmv12kPixelRemapper"]

@packed_struct
class ControlChannelWord:
    data_valid: unsigned(1)
    line_valid: unsigned(1)
    frame_valid: unsigned(1)
    fot: unsigned(1)
    integration_1: unsigned(1)
    integration_2: unsigned(1)


class Cmv12kPixelRemapper(Elaboratable):
    """Remaps one output port (if top and bottom outputs of the cmv12k are used, instantiate this twice!) to provide a linear output.

    Generally the CMV12000 sends its data out in bursts.
    The output format of the CMV12000 (and therefore the input for this core) depends on two factors:
    the number of lanes used and the presence of subsampling  / binning. The burst length is described by
    line_length / n_lanes_per_side so for the non line_lengh of 4096 and 16 lanes per side
    (the default beta configuration) this equals to a burst_length of 256.
    One Lane always sends adjacent pixels of bust_length out.

    """
    def __init__(self, top_lanes: List[Signal], bottom_lanes: List[Signal], control_lane: Signal):
        self.lanes = top_lanes
        self.bottom_lanes = bottom_lanes
        self.control_lane = control_lane
        assert len(top_lanes) == len(bottom_lanes)

        self.lines_to_buffer = 64 // (len(top_lanes) + len(bottom_lanes))
        self.n_bits = len(top_lanes[0])
        self.n_lanes_per_side = len(top_lanes)

        self.output = ImageStream(self.n_bits)

    def elaborate(self, platform):
        m = Module()

        cw = ControlChannelWord(self.control_lane)

        lane_top_streams_buffered = []
        for i, signal in enumerate(self.top_lanes):
            stream = BasicStream(self.n_bits)
            m.d.comb += stream.payload.eq(signal)
            m.d.comb += stream.valid.eq(cw.data_valid)

            fifo = m.submodules[f"fifo_top{i}"] = BufferedSyncStreamFIFO(stream, depth=128 * (self.lines_to_buffer + 1))
            lane_top_streams_buffered.append(fifo.output)

        return m



# An experiment to that checks the functionality of the axi writer
from nmigen import *
from nap import *


class Top(Elaboratable):
    def __init__(self):
        self.reset = ControlSignal()
        self.to_write = ControlSignal(reset=32 * 1024 * 1024)
        self.data_counter = StatusSignal(32)
        self.data_valid = ControlSignal()
        self.data_ready = StatusSignal()

    def elaborate(self, platform: ZynqSocPlatform):
        m = Module()

        platform.ps7.fck_domain(requested_frequency=200e6)
        m.d.comb += ResetSignal().eq(self.reset)

        stream = PacketizedStream(64)
        m.d.comb += self.data_ready.eq(stream.ready)
        m.d.comb += stream.valid.eq(self.data_valid)

        axi_writer = m.submodules.axi_writer = DramPacketRingbufferStreamWriter(stream, max_packet_size=0x1200000, n_buffers=4)

        with m.If(axi_writer.input.ready & axi_writer.input.valid):
            m.d.sync += self.data_counter.eq(self.data_counter + 1)
        m.d.comb += stream.payload.eq(Cat(self.data_counter, self.data_counter + 1000))

        return m


if __name__ == "__main__":
    cli(Top, runs_on=(MicroR2Platform, BetaPlatform, ZyboPlatform), possible_socs=(ZynqSocPlatform,))



# the 1x hdmi plugin module
# see: https://wiki.apertus.org/index.php/Beta_HDMI_Plugin_Module

from nmigen.build import Resource, Subsignal, Pins, PinsN, Attrs
from .plugin_connector import PluginDiffPair

__all__ = ["hdmi_plugin_connect"]


def hdmi_plugin_connect(platform, plugin_number):
    if "plugin_{}:gpio0".format(plugin_number) in platform._conn_pins:
        lowspeed_signals = [
            # i2c to read edid data from the monitor
            Subsignal("sda", Pins("lvds5_n", dir='io', conn=("plugin", plugin_number)), Attrs(IOSTANDARD="LVCMOS25")),
            Subsignal("scl", Pins("lvds5_p", dir='io', conn=("plugin", plugin_number)), Attrs(IOSTANDARD="LVCMOS25")),

            # hdmi plugin-module specific signals
            Subsignal("output_enable", PinsN("gpio6", dir='o', conn=("plugin", plugin_number)), Attrs(IOSTANDARD="LVCMOS33")),
            Subsignal("equalizer", Pins("gpio1 gpio4", dir='o', conn=("plugin", plugin_number)), Attrs(IOSTANDARD="LVCMOS33")),
            Subsignal("dcc_enable", Pins("gpio5", dir='o', conn=("plugin", plugin_number)), Attrs(IOSTANDARD="LVCMOS33")),
            Subsignal("vcc_enable", Pins("gpio7", dir='o', conn=("plugin", plugin_number)), Attrs(IOSTANDARD="LVCMOS33")),
            Subsignal("ddet", Pins("gpio3", dir='o', conn=("plugin", plugin_number)), Attrs(IOSTANDARD="LVCMOS33")),
            Subsignal("ihp", Pins("gpio2", dir='i', conn=("plugin", plugin_number)), Attrs(IOSTANDARD="LVCMOS33")),
        ]
    else:
        lowspeed_signals = []

    platform.add_resources([
        Resource("hdmi", plugin_number,
             Subsignal("clock", PluginDiffPair(platform, plugin_number, pin=3, dir='o', serdes=True), Attrs(IOSTANDARD="LVDS_25")),
             Subsignal("b", PluginDiffPair(platform, plugin_number, pin=2, dir='o', serdes=True), Attrs(IOSTANDARD="LVDS_25")),
             Subsignal("g", PluginDiffPair(platform, plugin_number, pin=1, dir='o', serdes=True), Attrs(IOSTANDARD="LVDS_25")),
             Subsignal("r", PluginDiffPair(platform, plugin_number, pin=0, dir='o', serdes=True), Attrs(IOSTANDARD="LVDS_25")),
             *lowspeed_signals
        )
    ])



# a helper for diagnosing ft601 performance

from nmigen import *

__all__ = ["FT601PerfDebug"]


class FT601PerfDebug(Elaboratable):
    def __init__(self, ft601_resource):
        self.ft_601_resource = ft601_resource
        self.burst_counter = Signal(16)
        self.idle_counter = Signal(16)

    def elaborate(self, platform):
        m = Module()

        ft = self.ft_601_resource

        m.domains += ClockDomain("ft")
        m.d.comb += ClockSignal("ft").eq(ft.clk)

        pause_cycles = 0
        in_transaction = Signal()
        pause_counter = Signal(range(pause_cycles + 1))

        with m.If((ft.txe & in_transaction) | (ft.txe & (pause_counter == pause_cycles))):  # we have space in the transmit fifo
            m.d.ft += self.burst_counter.eq(self.burst_counter + 1)
            m.d.comb += ft.write.eq(1)
            m.d.ft += self.idle_counter.eq(0)
            m.d.ft += in_transaction.eq(1)
            m.d.ft += pause_counter.eq(0)
        with m.Else():
            m.d.ft += self.burst_counter.eq(0)
            m.d.ft += self.idle_counter.eq(self.idle_counter + 1)
            m.d.ft += in_transaction.eq(0)
            with m.If(pause_counter < pause_cycles):
                m.d.ft += pause_counter.eq(pause_counter + 1)

        m.d.comb += ft.be.o.eq(0b1111)  # everything we write is valid
        m.d.comb += ft.oe.eq(0)  # we are driving the data bits all the time
        m.d.comb += ft.data.o.eq(Cat(self.burst_counter, self.idle_counter))

        return m



from os.path import join, dirname
from nmigen import Fragment, Signal
from ... import SocPlatform, Address, PeripheralsAggregator

__all__ = ["JTAGSocPlatform"]


class JTAGSocPlatform(SocPlatform):
    base_address = Address(address=0x0000_0000, bit_offset=0, bit_len=0xFFFF_FFFF * 8)
    pydriver_memory_accessor = open(join(dirname(__file__), "memory_accessor_openocd.py")).read()

    def __init__(self, platform):
        super().__init__(platform)

        self.jtag_active = Signal()
        self.jtag_debug_signals = Signal(32)

        def peripherals_connect_hook(platform, top_fragment: Fragment, sames):
            from nap import JTAGPeripheralConnector
            if platform.peripherals:
                aggregator = PeripheralsAggregator()
                for peripheral in platform.peripherals:
                    aggregator.add_peripheral(peripheral)

                jtag_controller = JTAGPeripheralConnector(aggregator)
                platform.to_inject_subfragments.append((jtag_controller, "jtag_controller"))

        self.prepare_hooks.append(peripherals_connect_hook)

    def pack_bitstream_fatbitstream(self, builder):
        builder.append_self_extracting_blob_from_file("{{name}}_sram.svf", "bitstream_jtag.svf")
        builder.append_command("openocd -f openocd.cfg -c 'svf -tap dut.tap -quiet -progress bitstream_jtag.svf; shutdown'\n")



import unittest
from nap import SimPlatform, BasicStream
from nap.stream import verify_stream_output_contract, LegalStreamSource, write_to_stream
from . import AxiEndpoint, answer_write_burst
from .stream_writer import AxiWriter, AxiWriterBurster, StreamPacketizer


class TestAxiWriter(unittest.TestCase):
    def test_basic(self):
        platform = SimPlatform()
        axi = AxiEndpoint(addr_bits=32, data_bits=64, lite=False, id_bits=12)

        data_stream = BasicStream(64)
        address_stream = BasicStream(32)

        write_sequence = [
            (0, 1),
            (0, 2),
            (0, 3),
            (0, 4),
            (1, 5),
            (2, 6),

            *[(a, a) for a in range(10, 1000, 8)],

            (1000, 1),
            (1000, 2),
            (1000, 3),
            (1000, 4),
            (1001, 5),
            (1002, 6),
        ]

        golden_memory = {}
        for addr, data in write_sequence:
            golden_memory[addr] = data

        dut = AxiWriter(address_stream, data_stream, axi)

        def write_address_process():
            for addr, data in write_sequence:
                yield from write_to_stream(address_stream, payload=addr)
        platform.add_process(write_address_process, "sync")

        def write_data_process():
            for addr, data in write_sequence:
                yield from write_to_stream(data_stream, payload=data)
        platform.add_process(write_data_process, "sync")

        def axi_answer_process():
            memory = {}
            while len(memory) < len(golden_memory):
                # print("m", len(memory), memory)
                written, accepted = (yield from answer_write_burst(axi))
                print("w", len(written), written)
                memory.update(written)
            self.assertEqual(golden_memory, memory)
        platform.add_process(axi_answer_process, "sync")

        platform.add_sim_clock("sync", 100e6)
        platform.sim(dut)

    def test_burster_address_output_stream_contract(self):
        data_input = BasicStream(64)
        address_input = BasicStream(32)

        dut = AxiWriterBurster(address_input, data_input)
        verify_stream_output_contract(
            dut, stream_output=dut.address_output,
            support_modules=(LegalStreamSource(data_input), LegalStreamSource(address_input))
        )

    def test_packetizer_output_stream_contract(self):
        length_input = BasicStream(4)
        data_input = BasicStream(64)

        dut = StreamPacketizer(length_input, data_input)
        verify_stream_output_contract(
            dut, support_modules=(LegalStreamSource(data_input), LegalStreamSource(length_input))
        )

    def test_burster_data_output_stream_contract(self):
        data_input = BasicStream(64)
        address_input = BasicStream(32)

        dut = AxiWriterBurster(address_input, data_input)
        verify_stream_output_contract(
            dut, stream_output=dut.data_output,
            support_modules=(LegalStreamSource(data_input), LegalStreamSource(address_input))
        )



from nmigen import *
from nmigen._unused import MustUse

from nap.soc.memorymap import Address
from nap.soc.peripheral import Response

__all__ = ["ControlSignal", "StatusSignal", "EventReg"]


class UncollectedCsrWarning(Warning):
    pass


class _Csr(MustUse):
    """a marker class to collect the registers easily"""
    _MustUse__warning = UncollectedCsrWarning
    _MustUse__silence = True
    _address = None


class ControlSignal(Signal, _Csr):
    """ Just a Signal. Indicator, that it is for controlling some parameter (i.e. can be written from the outside)
    Is mapped as a CSR in case the design is build with a SocPlatform.
    """

    def __init__(self, shape=None, *, address=None, read_strobe=None, write_strobe=None, src_loc_at=0, **kwargs):
        super().__init__(shape, src_loc_at=src_loc_at + 1, **kwargs)
        self.src_loc_at = src_loc_at
        self._shape = shape
        self._kwargs = kwargs

        self._address = Address.parse(address)
        self._write_strobe = write_strobe
        self._read_strobe = read_strobe


class StatusSignal(Signal, _Csr):
    """ Just a Signal. Indicator, that it is for communicating the state to the outside world (i.e. can be read but not written from the outside)
        Is mapped as a CSR in case the design is build with a SocPlatform.
    """

    def __init__(self, shape=None, *, address=None, read_strobe=None, src_loc_at=0, **kwargs):
        super().__init__(shape, src_loc_at=src_loc_at + 1, **kwargs)

        self._address = Address.parse(address)
        self._read_strobe = read_strobe


class EventReg(_Csr):  # TODO: bikeshed name
    """ A "magic" register, that doesnt have to be backed by a real register. Useful for implementing resets,
    fifo interfaces, ...
    The logic generated by the handle_read and handle_write hooks is part of the platform defined BusSlave and runs in its clockdomain.
    """

    def __init__(self, bits=None, address=None):
        super().__init__()
        assert address is not None or bits is not None
        self._bits = bits
        self._address = Address.parse(address)

        def handle_read(m, data, read_done):
            read_done(Response.ERR)

        self.handle_read = handle_read

        def handle_write(m, data, write_done):
            write_done(Response.ERR)

        self.handle_write = handle_write

    def __len__(self):
        if self._address is not None:
            return self._address.bit_len
        return self._bits



# Pass through gateware to program the usb3 plugin module on the Beta
# basically a nMigen adaption of http://vserver.13thfloor.at/Stuff/AXIOM/BETA/pass_jtag/
from nmigen import *
from nap import *


class Top(Elaboratable):
    def elaborate(self, platform):
        m = Module()

        usb3_plugin_connect(platform, "south", gpio=True, lvds=False, gpio_attrs=dict(IO_TYPE="LVCMOS33", PULLMODE="UP", DRIVE="4"))

        usb3 = platform.request("usb3_plugin", "south")
        pic_io = platform.request("pic_io")

        def connect(output, input, invert=False):
            m.d.comb += output.o.eq(~input.i if invert else input.i)
            m.d.comb += output.oe.eq(1)
            m.d.comb += input.oe.eq(0)

        connect(usb3.jtag.tdi, pic_io.sdo)
        connect(usb3.jtag.tck, pic_io.sck)
        connect(usb3.jtag_enb, pic_io.ss)
        connect(usb3.init, pic_io.initn, invert=True)
        connect(pic_io.done, usb3.done, invert=True)
        connect(pic_io.sdi, usb3.jtag.tdo)
        connect(usb3.jtag.tms, pic_io.sn)
        connect(usb3.program, pic_io.pb22b, invert=True)

        return m


if __name__ == "__main__":
    cli(Top, runs_on=(BetaRFWPlatform,))



from nmigen import *

__all__ = ["ISerdes8"]


class ISerdes8(Elaboratable):
    def __init__(self, input, ddr_domain, word_domain, invert=False):
        self.input = input
        self.output = Signal(8)
        self.bitslip = Signal()

        self.invert = invert
        self.ddr_domain = ddr_domain
        self.word_domain = word_domain

    def elaborate(self, platform):
        m = Module()

        iddr_output = Signal(8)
        m.d.comb += self.output.eq(iddr_output ^ Repl(self.invert, 8))
        m.submodules.iddr = Instance(
            "IDDRX4B",

            i_D=self.input,
            i_ECLK=ClockSignal(self.ddr_domain),
            i_SCLK=ClockSignal(self.word_domain),
            i_RST=ResetSignal(self.word_domain),
            i_ALIGNWD=self.bitslip,

            o_Q0=iddr_output[7],
            o_Q1=iddr_output[6],
            o_Q2=iddr_output[5],
            o_Q3=iddr_output[4],
            o_Q4=iddr_output[3],
            o_Q5=iddr_output[2],
            o_Q6=iddr_output[1],
            o_Q7=iddr_output[0],
        )

        return m


from nmigen import *

from nap.cores.peripherals.mmio_gpio import MmioGpio
from nap.soc.devicetree_overlay import devicetree_overlay
from nap.soc.soc_platform import SocPlatform

__all__ = ["BitbangI2c"]


class BitbangI2c(Elaboratable):
    def __init__(self, pins, name_suffix=""):
        self.devicetree_name = "bitbang_i2c" + name_suffix
        self.mmio_gpio = MmioGpio(pads=(pins.scl, pins.sda), name_suffix="_" + self.devicetree_name)

    def elaborate(self, platform: SocPlatform):
        m = Module()
        m.submodules.mmio_gpio = self.mmio_gpio

        overlay_content = """
            %overlay_name%: i2c@0 {
                compatible = "i2c-gpio";
                sda-gpios = <&%mmio_gpio% 1 6>;
                scl-gpios = <&%mmio_gpio% 0 6>;
                i2c-gpio,delay-us = <2>;        /* ~100 kHz */
                #address-cells = <1>;
                #size-cells = <0>;
            };
        """
        devicetree_overlay(platform, self.devicetree_name, overlay_content, {"mmio_gpio": self.mmio_gpio.devicetree_name})

        return m



from nmigen import *
from nap import nAny, iterator_with_if_elif
from . import ImageStream

__all__ = ["ImageSplitter", "ImageCombiner", "BlackLineGenerator"]


class ImageSplitter(Elaboratable):
    """Splits each image in four sub images. From each 4x4 pixel cluster each image receives one pixel. This can eg. be handy to decompose bayer data."""

    def __init__(self, input: ImageStream, width, height):
        self.input = input

        self.output_top_left = input.clone()
        self.output_top_right = input.clone()
        self.output_bottom_left = input.clone()
        self.output_bottom_right = input.clone()
        self.outputs = [self.output_top_left, self.output_top_right, self.output_bottom_left, self.output_bottom_right]
        self.output_shifts = [(0, 0), (1, 0), (0, 1), (1, 1)]

        self.width = width
        self.height = height

    def elaborate(self, platform):
        m = Module()

        input_transaction = Signal()
        m.d.comb += input_transaction.eq(self.input.ready & self.input.valid)

        x = Signal(16)
        y = Signal(16)
        with m.If(input_transaction):
            with m.If(~self.input.line_last):
                m.d.sync += x.eq(x + 1)
            with m.Else():
                m.d.sync += x.eq(0)
                m.d.sync += y.eq(y + 1)
            with m.If(self.input.frame_last):
                m.d.sync += y.eq(0)

        output_transaction = Signal()
        m.d.comb += output_transaction.eq(nAny(s.ready & s.valid for s in self.outputs))

        for cond, (output, shift) in iterator_with_if_elif(zip(self.outputs, self.output_shifts), m):
            with cond((x % 2 == shift[0]) & (y % 2 == shift[1])):
                m.d.comb += self.input.ready.eq(output.ready)
                m.d.comb += output.valid.eq(self.input.valid)
                m.d.comb += output.payload.eq(self.input.payload)
                # the last signals are used as first signals here and are later converted
                m.d.comb += output.line_last.eq((x // 2) == (self.width - 1) // 2)
                m.d.comb += output.frame_last.eq(((y // 2) == (self.height - 1) // 2) & output.line_last)

        return m


class ImageCombiner(Elaboratable):
    """Combines image streams to a larger image stream by either putting them side by side or interleaving them.
     May deadlock if the input streams are not enough buffered."""
    def __init__(self, *inputs: ImageStream, interleave=True, output_name=None):
        self.inputs = inputs
        assert all(input.payload.shape() == inputs[0].payload.shape() for input in inputs)
        self.output = inputs[0].clone(name=output_name)
        self.interleave = interleave

    def elaborate(self, platform):
        m = Module()

        output_transaction = Signal()
        m.d.comb += output_transaction.eq(self.output.ready & self.output.valid)

        with m.FSM():
            for i, input in enumerate(self.inputs):
                with m.State(str(i)):
                    m.d.comb += input.ready.eq(self.output.ready)
                    m.d.comb += self.output.valid.eq(input.valid)
                    m.d.comb += self.output.payload.eq(input.payload)
                    m.d.comb += self.output.line_last.eq(input.line_last & (i == len(self.inputs) - 1))
                    m.d.comb += self.output.frame_last.eq(input.frame_last & (i == len(self.inputs) - 1))
                    with m.If(self.interleave & output_transaction):
                        m.next = str((i + 1) % len(self.inputs))
                    with m.If((not self.interleave) & output_transaction & input.line_last):
                        m.next = str((i + 1) % len(self.inputs))

        return m


class BlackLineGenerator(Elaboratable):
    """generates a frame of infinite height and defined length. the generated frame is all black"""
    def __init__(self, payload_shape, width, black_value=0):
        self.output = ImageStream(payload_shape, name="black_lines_output")
        self.width = width
        self.black_value = black_value

    def elaborate(self, platform):
        m = Module()

        line_counter = Signal(range(self.width))
        m.d.comb += self.output.valid.eq(1)
        m.d.comb += self.output.payload.eq(self.black_value)
        with m.If(self.output.ready):
            with m.If(line_counter < self.width - 1):
                m.d.sync += line_counter.eq(line_counter + 1)
            with m.Else():
                m.d.sync += line_counter.eq(0)
                m.d.comb += self.output.line_last.eq(1)

        return m


from nmigen import *

from .axi_endpoint import AxiResponse as AxiResponse, AxiEndpoint
from nap.soc.peripheral import Response as BusSlaveResponse, Peripheral

__all__ = ["AxiLitePeripheralConnector"]


class AxiLitePeripheralConnector(Elaboratable):
    def __init__(self, peripheral: Peripheral, bundle_name="axi", timeout=1000):
        """
        A simple (low performance) axi lite `PeripheralConnector` for connecting `Peripheral`s to an AXI Lite Bus.
        :param peripheral: The peripheral which this controller should handle
        :param timeout: the timeout after which an unsuccessful read attempt should fail (useful for not hanging everything)
        """
        assert callable(peripheral.handle_read) and callable(peripheral.handle_write)
        self.peripheral = peripheral
        self.timeout = timeout

        self.axi = AxiEndpoint(addr_bits=32, data_bits=32, lite=True, name=bundle_name)

    def elaborate(self, platform):
        m = Module()

        address_range = self.peripheral.range()

        assert address_range is not None
        assert address_range.start < address_range.stop

        addr = Signal.like(self.axi.read_address.payload)
        read_out = Signal.like(self.axi.read_data.payload)
        resp_out = Signal.like(self.axi.read_data.resp)

        read_write_done = Signal()
        timeout_counter = Signal(range(self.timeout + 1))

        def read_write_done_callback(error):
            m.d.sync += read_write_done.eq(1)
            if error == BusSlaveResponse.ERR:
                m.d.sync += resp_out.eq(AxiResponse.SLVERR)
            else:
                m.d.sync += resp_out.eq(AxiResponse.OKAY)

        with m.FSM():
            with m.State("IDLE"):
                def in_range(signal):
                    return (signal >= address_range.start) & (signal < address_range.stop)

                with m.If(self.axi.read_address.valid & in_range(self.axi.read_address.payload)):
                    m.next = "FETCH_READ_ADDRESS"
                with m.Elif(self.axi.write_address.valid & in_range(self.axi.write_address.payload)):
                    m.next = "FETCH_WRITE_ADDRESS"

            with m.State("FETCH_READ_ADDRESS"):
                m.d.sync += addr.eq(self.axi.read_address.payload - address_range.start)
                m.d.comb += self.axi.read_address.ready.eq(1)
                m.next = "READ"
                m.d.sync += timeout_counter.eq(0)
            with m.State("READ"):
                # Only write read and resp if we are in the READ state. Otherwise all bits are '0'.
                # This allows us to simply or the data output of multiple axi slaves together.
                m.d.sync += timeout_counter.eq(timeout_counter + 1)
                self.peripheral.handle_read(m, addr, read_out, read_write_done_callback)
                with m.If(read_write_done):
                    m.d.sync += read_write_done.eq(0)
                    m.next = "READ_DONE"
                with m.If(timeout_counter == self.timeout):
                    m.d.sync += read_write_done.eq(1)
                    m.d.sync += resp_out.eq(AxiResponse.DECERR)
            with m.State("READ_DONE"):
                m.d.comb += self.axi.read_data.payload.eq(read_out)
                m.d.comb += self.axi.read_data.resp.eq(resp_out)
                m.d.comb += self.axi.read_data.valid.eq(1)
                with m.If(self.axi.read_data.ready):
                    m.next = "IDLE"

            with m.State("FETCH_WRITE_ADDRESS"):
                m.d.sync += addr.eq(self.axi.write_address.payload - address_range.start)
                m.d.comb += self.axi.write_address.ready.eq(1)
                m.next = "WRITE"
                m.d.sync += timeout_counter.eq(0)
            with m.State("WRITE"):
                with m.If(self.axi.write_data.valid):
                    self.peripheral.handle_write(m, addr, self.axi.write_data.payload, read_write_done_callback)
                    m.d.sync += timeout_counter.eq(timeout_counter + 1)
                    with m.If(read_write_done):
                        m.d.sync += read_write_done.eq(0)
                        m.next = "WRITE_DONE"
                    with m.If(timeout_counter == self.timeout):
                        m.d.sync += read_write_done.eq(1)
                        m.d.sync += resp_out.eq(AxiResponse.DECERR)
            with m.State("WRITE_DONE"):
                m.d.comb += self.axi.write_response.resp.eq(resp_out)
                m.d.comb += self.axi.write_response.valid.eq(1)
                m.d.comb += self.axi.write_data.ready.eq(1)
                with m.If(self.axi.write_response.ready):
                    m.next = "IDLE"

        return m



from .axi_endpoint import *
from .full_to_lite import *
from .interconnect import *
from .peripheral_connector import *
from .sim_util import *
from .stream_reader import *
from .stream_writer import *
from .zynq_util import *



# TODO: add tests (a lot of them)

import re
from dataclasses import dataclass
from math import ceil
from typing import List

from nmigen._unused import MustUse

from .pydriver.drivermethod import DriverMethod

__all__ = ["Address", "MemoryMap"]


class Address:
    @staticmethod
    def parse(obj, bus_word_width=32):
        """
        Parse an address object from either a string or pass an address object or none through.
        :param obj: either a string in the format 0xABCD[optional_python:slice] or an Address object or None
        :param bus_word_width: the native word width of the underlying bus in bits.
        :return: an address object or None
        """
        if isinstance(obj, Address):
            return obj
        elif isinstance(obj, str):
            addr, start_bit, stop_bit = [
                int(s, 0) if s else None for s in
                re.match("(0x[0-9a-fA-F]+)\\[?(\\d+)?:?(\\d+)?\\]?", obj).groups()
            ]
            if (start_bit is not None) and (stop_bit is not None):
                return Address(addr, start_bit, stop_bit - start_bit)
            elif start_bit is not None:
                return Address(addr, start_bit, 1)
            else:
                return Address(addr, 0, bus_word_width)
        elif obj is None:
            return None
        else:
            raise ValueError("constructing addresses from type {!r} is not supported".format(type(obj)))

    def __init__(self, address, bit_offset, bit_len=None, bus_word_width=32):
        """
        An address type, that can either represent only a base adress or a base adress and a length.
        :param address: the base address in bytes
        :param bit_offset: the offset of the start of the address unit in bits relative to the start byte. must be positive.
        :param bit_len: the length of the address unit in bits.
        :param bus_word_width: the native word width of the underlying bus in bits.
        """
        assert bus_word_width % 8 == 0
        self.bus_word_width = bus_word_width

        assert bit_offset <= bus_word_width
        self.bit_offset = bit_offset

        assert address % int(bus_word_width / 8) == 0
        self.address = address

        self.bit_len = bit_len

    def __repr__(self):
        if self.bit_len is not None:
            return "0x{:02X}[{}:{}]".format(self.address, self.bit_offset, self.bit_offset + self.bit_len)
        else:
            return "0x{:02X}[{}:]".format(self.address, self.bit_offset)

    def collides(self, other_address):
        if (self.bit_len is None) or (other_address.bit_len is None):
            raise ValueError("collision detection is impossible with addresses that dont have a length specified")
        self_start = self.address * 8
        self_stop = self_start + self.bit_len

        other_start = other_address.address * 8
        other_stop = other_start + other_address.bit_len

        if self_start <= other_start:
            return self_stop > other_start
        else:
            return other_stop > self_start

    def bits_of_word(self, word_address: int):
        """
        If the Address is part of the addressed word it returns a tuple of the range of own addressed bits and of the
        bits of the word. otherwise it returns None.
        :param word_address: the base address of the word. this is not an Address object but rather an int because we
        need only the base address.
        :return a tuple of (word_range, signal_range)
        """
        word = Address(word_address, 0, self.bus_word_width)
        if word.collides(self):
            address_shift = word_address - self.address
            if address_shift == 0:  # we are in the start word
                stop = max(self.bit_len - self.bit_offset, word.bit_len - self.bit_offset)
                word_range = range(self.bit_offset, stop)
                signal_range = range(0, stop)
            else:  # we are in some subsequent word
                stop = max((self.bit_len + self.bit_offset) - self.bus_word_width, self.bus_word_width)
                word_range = range(0, stop)
                signal_start = address_shift * 8 - self.bit_offset
                signal_range = range(signal_start, signal_start + stop)
            return word_range, signal_range
        else:
            return None

    def translate(self, other):
        """
        Translate an other address by the current address
        :param other: the address to translate
        :return: the translated address
        """
        assert isinstance(other, Address)
        assert self.bit_offset == 0
        assert other.address * 8 + other.bit_offset + other.bit_len <= self.bit_len
        return Address(self.address + other.address, other.bit_offset, other.bit_len)

    def range(self):
        """
        Return a range(start_byte, end_byte)
        """
        return range(self.address, self.address + ceil((self.bit_offset + self.bit_len) / 8))


@dataclass
class MemoryMapRow:
    name: str
    address: Address
    writable: bool
    obj: object


class UnusedMemoryMap(Warning):
    pass


class MemoryMap(MustUse):
    def __init__(self, place_at=None, parent=None, top=False, bus_word_width=32):
        """
        A memorymap which can store & allocate (hierarchical) address -> resource mappings
        :param place_at: the optional preference of where the memorymap should be located
        :param bus_word_width: the native word width of the underlying bus in bits.
        """
        assert bus_word_width % 8 == 0
        self.bus_word_width = bus_word_width
        self.place_at: Address = place_at

        self.is_top = top
        self._parent = parent
        self._inlined_offset = None

        self.entries: List[MemoryMapRow] = []
        self.aliases = {}
        self.driver_methods = {}
        self.frozen = False
        self._MustUse__warning = UnusedMemoryMap

    def __repr__(self):
        return "<Memorymap top={} byte_len={} at {}>".format(self.is_top, self.byte_len, hex(id(self)))

    @property
    def is_empty(self):
        return not self.entries

    @property
    def _MustUse__silence(self):
        return self.is_top or self._parent is not None or self.is_empty

    @property
    def top_memorymap(self):
        if self.is_top:
            return self
        elif self._parent is not None:
            return self._parent.top_memorymap
        else:
            raise ValueError("self is not the toplevel memorymap and is not assigned to one")

    @property
    def path(self):
        if self.is_top:
            return ()
        elif self._parent is not None:
            my_row: MemoryMapRow = next(row for row in self._parent.subranges if row.obj == self)
            return (*self._parent.path, my_row.name)
        else:
            raise ValueError("self is not the toplevel memorymap and is not assigned to one")

    @property
    def was_inlined(self):
        return self._inlined_offset is not None

    @property
    def bus_word_width_bytes(self):
        return int(self.bus_word_width / 8)

    @property
    def subranges(self):
        return [row for row in self.entries if isinstance(row.obj, MemoryMap)]

    @property
    def direct_children(self):
        return [row for row in self.entries if not isinstance(row.obj, MemoryMap)]

    @property
    def byte_len(self) -> int:
        """
        Calculate the size (based on the resource with the highest address part) of the memorymap
        :return: the size of the memorymap in bytes (aligned to the bus word width)
        """
        if not self.entries:
            return 0
        real_size = max(
            row.address.address + ceil((row.address.bit_offset + row.address.bit_len) / 8)
            for row in self.entries
        )
        # automatically round up to the next word with self.access_with
        return int(ceil(real_size / self.bus_word_width_bytes) * self.bus_word_width_bytes)

    @property
    def direct_children_byte_len(self):
        """
        Calculate the size (based on the _normal_ resource with the highest address part) of the memorymap
        :return: the size of the memorymap in bytes (aligned to the bus word width)
        """
        if not self.direct_children:
            return 0
        real_size = max(
            row.address.address + ceil((row.address.bit_offset + row.address.bit_len) / 8)
            for row in self.direct_children
        )
        # automatically round up to the next word with self.access_with
        return int(ceil(real_size / self.bus_word_width_bytes) * self.bus_word_width_bytes)

    @property
    def absolute_range_of_direct_children(self):
        own_offset = self.own_offset
        own_offset.bit_len = self.direct_children_byte_len * 8
        return own_offset

    @property
    def own_offset(self) -> Address:
        if self.is_empty:
            return Address(0, 0, 0)
        if not self.is_top:
            if self._parent and self._inlined_offset:
                return self._parent.own_offset.translate(self._inlined_offset)
            elif self._parent:
                own_row_candidates = [row for row in self._parent.subranges if row.obj == self]
                assert len(own_row_candidates) == 1
                return self._parent.own_offset.translate(own_row_candidates[0].address)
            else:
                raise ValueError("the location of the memorymap cant be determined. "
                                 "self is not the toplevel memorymap and is not assigned to one")
        else:
            if self.place_at:
                return Address(self.place_at.address, 0, self.byte_len * 8)
            else:
                return Address(0, 0, self.byte_len * 8)

    def is_free(self, check_address):
        """
        Check if the provided address is free.
        :param check_address: the address to check.
        :return: A boolean indicating if the address is free
        """
        for row in self.entries:
            if check_address.collides(row.address):
                return False
        return True

    def allocate(self, name, writable, bits=None, address=None, obj=None):
        """
        Allocate / add a resource to the memorymap.
        :param name: the name of the resource
        :param writable: a boolean indicating if the resource is writable
        :param bits: the width of the peripheral
        :param address: the address to place the resource to (optional)
        :param obj: an object to attach to that memory location. In the case of hierarchical memorymaps,
        this is the sub memorymap
        :return: the address of the resource
        """
        assert not self.frozen
        assert not any(row.name == name for row in self.entries), name
        assert bits is not None or address is not None
        if address:
            assert ((bits is None) or (
                    address.bit_len is None)) and bits != address.bit_len, "conflicting length information"
        if not address:
            address = Address(self.byte_len, 0, bits, self.bus_word_width)
        elif bits:
            address.bit_len = bits
        assert address.bit_len
        if not self.is_free(address):
            raise ValueError("address {!r} is not free".format(address, bits))
        self.entries.append(MemoryMapRow(name, address, writable, obj))
        return address

    def add_alias(self, name, obj):
        """
        Adds an alias to a resource somewhere else in the hierarchy.
        :param name: the name of the resource
        :param obj: the object with which one can find the other resource
        """
        self.aliases[name] = obj

    def add_driver_method(self, name, drivermethod):
        assert isinstance(drivermethod, DriverMethod)
        self.driver_methods[name] = drivermethod

    def _added_to(self, parent, inlined_offset=None):
        self.frozen = True
        self._parent = parent
        self._inlined_offset = inlined_offset
        return self

    def allocate_subrange(self, subrange, name=None, place_at=None):
        """
        Add a sub-memorymap to the Memorymap.
        :type subrange: MemoryMap
        :param name: the name of the subrange. If None the subrange be inlined into this memorymap and no hierarchy will be created.
        :param subrange: The subrange to add
        :param place_at: The desired location of the memorymap fragment. Overrides the subranges preference.
        :return: The address where the subrange is placed
        """
        if subrange.byte_len == 0:
            subrange._added_to(None)
            return  # we do not need to do anything with subranges that are empty
        if not place_at:
            place_at = self.place_at
        assert self.place_at is None
        if name is None:  # inline the memorymap
            place_to = Address(self.byte_len, 0, subrange.byte_len * 8)
            for row in subrange._added_to(self, inlined_offset=place_to).entries:
                self.allocate(row.name, row.writable, address=place_to.translate(row.address), obj=row.obj)
            return place_to
        else:  # add the memorymap as a regular resource. this is later interpreted as hierarchical memorymaps
            return self.allocate(name, True, bits=subrange.byte_len * 8, address=place_at, obj=subrange._added_to(self))

    def find_recursive(self, obj, go_up=False):
        """
        Searches recursively for the given object and returns the associated address.
        If the object is not found, None is returned
        :param go_up: Determine if we should search upwards in the hierarchy if we are not the top memorymap
        :rtype: Address
        :param obj: the object to look for
        """

        if go_up:
            return self.top_memorymap.find_recursive(obj, go_up=False)

        for row in self.direct_children:
            if row.obj is obj or row is obj:
                return self.own_offset.translate(row.address)

        for row in self.subranges:
            sub_result = row.obj.find_recursive(obj)
            if sub_result is not None:
                return sub_result

        return None

    @property
    def flattened(self):
        to_return = {}
        for row in self.direct_children:
            to_return[(*self.path, row.name)] = self.find_recursive(row, go_up=True)
        for name, obj in self.aliases.items():
            to_return[(*self.path, name)] = self.find_recursive(obj, go_up=True)
        for name, method in self.driver_methods.items():
            to_return[(*self.path, name)] = method
        for row in self.subranges:
            to_return.update(row.obj.flattened)

        return to_return



# the usb3 plugin module plugin module resource
# (not the platform definition for building gateware for the fpga _on_ the plugin module itself)
# see: https://wiki.apertus.org/index.php/1x_USB_3.0_Plugin_Module

from nmigen.build import *
from .plugin_connector import PluginDiffPair

__all__ = ["usb3_plugin_connect"]


def usb3_plugin_connect(platform, plugin_number, gpio=True, lvds=True, gpio_attrs=dict(IOSTANDARD="LVCMOS25")):
    if gpio:
        gpio_signals = [
            Subsignal(
                "jtag",
                Subsignal("tms", Pins("gpio0", dir="io", conn=("plugin", plugin_number)), Attrs(**gpio_attrs)),
                Subsignal("tck", Pins("gpio1", dir="io", conn=("plugin", plugin_number)), Attrs(**gpio_attrs)),
                Subsignal("tdi", Pins("gpio2", dir="io", conn=("plugin", plugin_number)), Attrs(**gpio_attrs)),
                Subsignal("tdo", Pins("gpio3", dir="io", conn=("plugin", plugin_number)), Attrs(**gpio_attrs)),
            ),
            Subsignal("jtag_enb", Pins("gpio4", dir="io", conn=("plugin", plugin_number)), Attrs(**gpio_attrs)),
            Subsignal("program", PinsN("gpio5", dir="io", conn=("plugin", plugin_number)), Attrs(**gpio_attrs)),
            Subsignal("init", PinsN("gpio6", dir="io", conn=("plugin", plugin_number)), Attrs(**gpio_attrs)),
            Subsignal("done", PinsN("gpio7", dir="io", conn=("plugin", plugin_number)), Attrs(**gpio_attrs)),
        ]
    else:
        gpio_signals = []

    if lvds:
        lvds_signals = [
            Subsignal(
                "lvds",
                Subsignal("valid", PluginDiffPair(platform, plugin_number, 0, dir='o', serdes=True), Attrs(IOSTANDARD="LVDS_25")),
                Subsignal("lane0", PluginDiffPair(platform, plugin_number, 1, dir='o', serdes=True), Attrs(IOSTANDARD="LVDS_25")),
                Subsignal("lane1", PluginDiffPair(platform, plugin_number, 2, dir='o', serdes=True), Attrs(IOSTANDARD="LVDS_25")),
                Subsignal("lane2", PluginDiffPair(platform, plugin_number, 3, dir='o', serdes=True), Attrs(IOSTANDARD="LVDS_25")),
                Subsignal("lane3", PluginDiffPair(platform, plugin_number, 4, dir='o', serdes=True), Attrs(IOSTANDARD="LVDS_25")),
                Subsignal("clk_word", PluginDiffPair(platform, plugin_number, 5, dir='o', serdes=True), Attrs(IOSTANDARD="LVDS_25")),
            ),
        ]
    else:
        lvds_signals = []

    platform.add_resources([
        Resource(
            "usb3_plugin", plugin_number,
            *lvds_signals,
            *gpio_signals,
        )
    ])



import mmap
import os
import struct
from math import ceil


class DevMemAccessor:
    word = 4
    mask = ~(word - 1)

    def __init__(self, base_addr=0x4000_0000, bytes=None, filename='/dev/mem'):
        if bytes is None:
            bytes = mmap.PAGESIZE

        assert (base_addr % mmap.PAGESIZE) == 0

        bytes = int(ceil(bytes / mmap.PAGESIZE) * mmap.PAGESIZE)

        self.base = base_addr

        self.f = os.open(filename, os.O_RDWR | os.O_SYNC)
        self.mem = mmap.mmap(self.f, bytes, mmap.MAP_SHARED, mmap.PROT_READ | mmap.PROT_WRITE, offset=base_addr)

    def __del__(self):
        os.close(self.f)

    def read(self, offset):
        self.mem.seek(offset)
        return struct.unpack('I', self.mem.read(4))[0]

    def write(self, offset, to_write):
        self.mem.seek(offset)
        return self.mem.write(struct.pack('I', to_write))


MemoryAccessor = DevMemAccessor



# Provides a linux framebuffer via HDMI

from nmigen import *
from nap import *


class Top(Elaboratable):
    def __init__(self):
        self.width = 1280
        self.height = 720

    def elaborate(self, platform: ZynqSocPlatform):
        if not isinstance(platform, ZyboPlatform):
            hdmi_plugin_connect(platform, "north")

        m = Module()

        cpu_writer = m.submodules.cpu_writer = DramPacketRingbufferCpuWriter(
            max_packet_size=0x1000000, n_buffers=1,
            default_packet_size=self.width * self.height * 4
        )

        platform.ps7.fck_domain(100e6, "axi_hp")

        p = Pipeline(m, start_domain="axi_hp")
        p += DramPacketRingbufferStreamReader(cpu_writer)
        p += SimpleStreamGearbox(p.output, target_width=32)
        p += StreamResizer(p.output, target_width=24)
        p += PacketizedStream2ImageStream(p.output, width=self.width)
        p += BufferedAsyncStreamFIFO(p.output, depth=16 * 1024, o_domain="pix")

        hdmi = platform.request("hdmi", "north")
        p += HdmiStreamSink(p.output, hdmi, generate_modeline(self.width, self.height, 30), pix_domain="pix")

        m.submodules.clocking_debug = ClockingDebug("pix", "pix_5x", "axi_hp")

        overlay_content = """
            %overlay_name%: framebuffer@%address% {
                compatible = "simple-framebuffer";
                reg = <0x%address% (%width% * %height% * 4)>;
                width = <%width%>;
                height = <%height%>;
                stride = <(%width% * 4)>;
                format = "a8b8g8r8";
            };
        """
        devicetree_overlay(platform, "framebuffer", overlay_content, {
            "width": str(self.width),
            "height": str(self.height),
            "address": "{:x}".format(cpu_writer.buffer_base_list[0]),
        })

        return m


if __name__ == "__main__":
    cli(Top, runs_on=(MicroR2Platform, BetaPlatform, ZyboPlatform), possible_socs=(ZynqSocPlatform, ))



from collections import defaultdict
from typing import List
from nmigen import *
from nap import Stream, DOWNWARDS, nAll
from . import StreamBuffer

__all__ = ["StreamTee", "StreamCombiner"]


class StreamTee(Elaboratable):
    """Routes a single stream input to multiple outputs (like a tee shaped piece of pipe).
    To be able to fulfil the stream contract, every output has a StreamBuffer.
    """

    def __init__(self, input: Stream):
        self.input = input
        self.outputs: List[Stream] = []
        self.m = Module()

    def get_output(self):
        n = len(self.outputs)
        output_before_buffer = self.input.clone(name=f"output{n}_before_fifo")
        self.outputs.append(output_before_buffer)
        output_fifo = StreamBuffer(output_before_buffer)
        self.m.submodules[f"output{n}_fifo"] = output_fifo
        return output_fifo.output

    def elaborate(self, platform):
        m = self.m

        m.d.comb += self.input.ready.eq(nAll(output.ready for output in self.outputs))
        for output in self.outputs:
            m.d.comb += output.valid.eq(self.input.valid & self.input.ready)
            for k in self.input.payload_signals.keys():
                u, d = self.input.payload_signals[k], output.payload_signals[k]
                m.d.comb += d.eq(u)

        return m


class StreamCombiner(Elaboratable):
    def __init__(self, *inputs: Stream, **merge_plan):

        payload_signals = defaultdict(list)
        for input in inputs:
            for k, signal in input.payload_signals.items():
                payload_signals[k].append(signal)

        payload_expressions = {}
        for k, signals in payload_signals.items():
            if "exclude_{}".format(k) in merge_plan:
                continue
            elif "merge_{}".format(k) in merge_plan:
                payload_expressions[k] = Cat(signals)
            else:
                payload_expressions[k] = signals[0]

        class CombinedStream(Stream):
            def __init__(self, name=None, src_loc_at=1):
                super().__init__(name=name, src_loc_at=1 + src_loc_at)
                for k, expr in payload_expressions.items():
                    setattr(self, k, Signal(len(expr)) @ DOWNWARDS)
        self.payload_expressions = payload_expressions

        self.inputs = inputs
        self.output = CombinedStream()

    def elaborate(self, platform):
        m = Module()

        m.d.comb += self.output.valid.eq(nAll(input.valid for input in self.inputs))
        for input in self.inputs:
            m.d.comb += input.ready.eq(self.output.ready)
        for k, expr in self.payload_expressions.items():
            m.d.comb += self.output[k].eq(expr)

        return m

